#  Copyright (c) 2022-2023.
#  ProrokLab (https://www.proroklab.org/)
#  All rights reserved.
import math
from typing import Dict

import torch
from torch import Tensor

from vmas import render_interactively
from vmas.simulator.core import Agent, Box, Landmark, Sphere, World, Line
from vmas.simulator.joints import Joint
from vmas.simulator.scenario import BaseScenario
from vmas.simulator.utils import Color, Y, X
from vmas.simulator.velocity_controller import VelocityController


def angle_to_vector(angle):
    return torch.cat([torch.cos(angle), torch.sin(angle)], dim=1)


def get_line_angle_0_90(rot: Tensor):
    angle = torch.abs(rot) % torch.pi
    other_angle = torch.pi - angle
    return torch.minimum(angle, other_angle)


def get_line_angle_0_180(rot):
    angle = rot % torch.pi
    return angle


def get_line_angle_dist_0_360(angle, goal):
    angle = angle_to_vector(angle)
    goal = angle_to_vector(goal)
    return -torch.einsum("bs,bs->b", angle, goal)


def get_line_angle_dist_0_180(angle, goal):
    angle = get_line_angle_0_180(angle)
    goal = get_line_angle_0_180(goal)
    return torch.minimum(
        (angle - goal).abs(),
        torch.minimum(
            (angle - (goal - torch.pi)).abs(), ((angle - torch.pi) - goal).abs()
        ),
    ).squeeze(-1)


class Scenario(BaseScenario):
    def make_world(self, batch_dim: int, device: torch.device, **kwargs):
        self.fixed_passage = kwargs.get("fixed_passage", False)
        self.joint_length = kwargs.get("joint_length", 0.52)
        self.random_start_angle = kwargs.get("random_start_angle", False)
        self.random_goal_angle = kwargs.get("random_goal_angle", False)
        self.observe_joint_angle = kwargs.get("observe_joint_angle", False)
        self.joint_angle_obs_noise = kwargs.get("joint_angle_obs_noise", 0.0)
        self.asym_package = kwargs.get("asym_package", False)
        self.mass_ratio = kwargs.get("mass_ratio", 1)
        self.mass_position = kwargs.get("mass_position", 0.75)
        self.max_speed_1 = kwargs.get("max_speed_1", None)  # 0.1
        self.pos_shaping_factor = kwargs.get("pos_shaping_factor", 1)
        self.rot_shaping_factor = kwargs.get("rot_shaping_factor", 1)
        self.collision_reward = kwargs.get("collision_reward", 0)
        self.energy_reward_coeff = kwargs.get("energy_reward_coeff", 0)
        self.obs_noise = kwargs.get("obs_noise", 0.0)

        self.plot_grid = False

        # Make world
        world = World(
            batch_dim,
            device,
            x_semidim=1,
            y_semidim=1,
            substeps=5 if not self.asym_package else 10,
            joint_force=700 if self.asym_package else 400,
            collision_force=2500 if self.asym_package else 1500,
            drag=0.25 if not self.asym_package else 0.15,
        )

        if not self.observe_joint_angle:
            assert self.joint_angle_obs_noise == 0

        self.n_agents = 2

        self.middle_angle = torch.zeros((world.batch_dim, 1), device=world.device)

        self.agent_radius = 0.03333
        self.agent_radius_2 = 3 * self.agent_radius
        self.mass_radius = self.agent_radius * (2 / 3)
        self.passage_width = 0.2
        self.passage_length = 0.1476
        self.scenario_length = 2 + 2 * self.agent_radius
        self.n_boxes = int(self.scenario_length // self.passage_length)
        self.min_collision_distance = 0.005

        cotnroller_params = [2.0, 10, 0.00001]

        # Add agents
        agent = Agent(
            name="agent 0",
            shape=Sphere(self.agent_radius),
            u_range=0.5,
            obs_noise=self.obs_noise,
            render_action=True,
            f_range=10,
        )
        agent.controller = VelocityController(
            agent, world, cotnroller_params, "standard"
        )
        world.add_agent(agent)
        agent = Agent(
            name="agent 1",
            shape=Sphere(self.agent_radius_2),
            u_range=0.5,
            mass=1 if self.asym_package else self.mass_ratio,
            max_speed=self.max_speed_1,
            obs_noise=self.obs_noise,
            render_action=True,
            f_range=10,
        )
        agent.controller = VelocityController(
            agent, world, cotnroller_params, "standard"
        )
        world.add_agent(agent)

        self.joint = Joint(
            world.agents[0],
            world.agents[1],
            anchor_a=(0, 0),
            anchor_b=(0, 0),
            dist=self.joint_length,
            rotate_a=True,
            rotate_b=True,
            collidable=False,
            width=0,
            mass=1,
        )
        world.add_joint(self.joint)

        if self.asym_package:

            def mass_collision_filter(e):
                return not isinstance(e.shape, Sphere)

            self.mass = Landmark(
                name="mass",
                shape=Sphere(radius=self.mass_radius),
                collide=True,
                movable=True,
                color=Color.BLACK,
                mass=self.mass_ratio,
                collision_filter=mass_collision_filter,
            )

            world.add_landmark(self.mass)

            joint = Joint(
                self.mass,
                self.joint.landmark,
                anchor_a=(0, 0),
                anchor_b=(self.mass_position, 0),
                dist=0,
                rotate_a=True,
                rotate_b=True,
            )
            world.add_joint(joint)

        self.goal = Landmark(
            name="joint_goal",
            shape=Line(length=self.joint_length),
            collide=False,
            color=Color.GREEN,
        )
        world.add_landmark(self.goal)

        self.walls = []
        for i in range(4):
            wall = Landmark(
                name=f"wall {i}",
                collide=True,
                shape=Line(length=2 + self.agent_radius * 2),
                color=Color.BLACK,
            )
            world.add_landmark(wall)
            self.walls.append(wall)

        self.create_passage_map(world)

        self.pos_rew = torch.zeros(batch_dim, device=device)
        self.rot_rew = self.pos_rew.clone()
        self.collision_rew = self.pos_rew.clone()
        self.energy_rew = self.pos_rew.clone()
        self.all_passed = torch.full((batch_dim,), False, device=device)

        return world

    def reset_world_at(self, env_index: int = None):
        start_angle = torch.rand(
            (1, 1) if env_index is not None else (self.world.batch_dim, 1),
            device=self.world.device,
        )
        start_angle[start_angle >= 0.5] = torch.pi / 2
        start_angle[start_angle < 0.5] = -torch.pi / 2

        goal_angle = torch.zeros(
            (1, 1) if env_index is not None else (self.world.batch_dim, 1),
            device=self.world.device,
            dtype=torch.float32,
        ).uniform_(
            -torch.pi / 2 if self.random_goal_angle else torch.pi,
            torch.pi / 2 if self.random_goal_angle else torch.pi,
        )

        bigger_radius = max(self.agent_radius, self.agent_radius_2)

        start_delta_x = (self.joint_length / 2) * torch.cos(start_angle)
        start_delta_x_abs = start_delta_x.abs()
        min_x_start = -self.world.x_semidim + (bigger_radius + start_delta_x_abs)
        max_x_start = self.world.x_semidim - (bigger_radius + start_delta_x_abs)
        start_delta_y = (self.joint_length / 2) * torch.sin(start_angle)
        start_delta_y_abs = start_delta_y.abs()
        min_y_start = -self.world.y_semidim + (bigger_radius + start_delta_y_abs)
        max_y_start = -2 * bigger_radius - self.passage_width / 2 - start_delta_y_abs

        goal_delta_x = (self.joint_length / 2) * torch.cos(goal_angle)
        goal_delta_x_abs = goal_delta_x.abs()
        min_x_goal = -self.world.x_semidim + (bigger_radius + goal_delta_x_abs)
        max_x_goal = self.world.x_semidim - (bigger_radius + goal_delta_x_abs)
        goal_delta_y = (self.joint_length / 2) * torch.sin(goal_angle)
        goal_delta_y_abs = goal_delta_y.abs()
        min_y_goal = 2 * bigger_radius + self.passage_width / 2 + goal_delta_y_abs
        max_y_goal = self.world.y_semidim - (bigger_radius + goal_delta_y_abs)

        joint_pos = torch.cat(
            [
                (min_x_start - max_x_start)
                * torch.rand(
                    (1, 1) if env_index is not None else (self.world.batch_dim, 1),
                    device=self.world.device,
                    dtype=torch.float32,
                )
                + max_x_start,
                (min_y_start - max_y_start)
                * torch.rand(
                    (1, 1) if env_index is not None else (self.world.batch_dim, 1),
                    device=self.world.device,
                    dtype=torch.float32,
                )
                + max_y_start,
            ],
            dim=1,
        )
        goal_pos = torch.cat(
            [
                (min_x_goal - max_x_goal)
                * torch.rand(
                    (1, 1) if env_index is not None else (self.world.batch_dim, 1),
                    device=self.world.device,
                    dtype=torch.float32,
                )
                + max_x_goal,
                (min_y_goal - max_y_goal)
                * torch.rand(
                    (1, 1) if env_index is not None else (self.world.batch_dim, 1),
                    device=self.world.device,
                    dtype=torch.float32,
                )
                + max_y_goal,
            ],
            dim=1,
        )

        self.goal.set_pos(
            goal_pos,
            batch_index=env_index,
        )
        self.goal.set_rot(goal_angle, batch_index=env_index)

        agents = self.world.agents
        for i, agent in enumerate(agents):
            agent.controller.reset(env_index)
            if i == 0:
                agent.set_pos(
                    joint_pos - torch.cat([start_delta_x, start_delta_y], dim=1),
                    batch_index=env_index,
                )
            else:
                agent.set_pos(
                    joint_pos + torch.cat([start_delta_x, start_delta_y], dim=1),
                    batch_index=env_index,
                )

        if self.asym_package:
            self.mass.set_pos(
                joint_pos
                + self.mass_position
                * torch.cat([start_delta_x, start_delta_y], dim=1)
                * (1 if agents[0] == self.world.agents[0] else -1),
                batch_index=env_index,
            )

        self.spawn_passage_map(env_index)

        self.spawn_walls(env_index)

        if env_index is None:
            self.passed = torch.zeros((self.world.batch_dim,), device=self.world.device)

            self.joint.pos_shaping_pre = (
                torch.linalg.vector_norm(
                    self.joint.landmark.state.pos - self.pass_center, dim=1
                )
                * self.pos_shaping_factor
            )
            self.joint.pos_shaping_post = (
                torch.linalg.vector_norm(
                    self.joint.landmark.state.pos - self.goal.state.pos, dim=1
                )
                * self.pos_shaping_factor
            )

            self.joint.rot_shaping_pre = (
                get_line_angle_dist_0_360(
                    self.joint.landmark.state.rot, self.middle_angle
                )
                * self.rot_shaping_factor
            )
            # self.joint.rot_shaping_post = (
            #     get_line_angle_dist_0_180(
            #         self.joint.landmark.state.rot, self.goal.state.rot
            #     )
            #     * self.rot_shaping_factor
            # )

        else:
            self.passed[env_index] = 0

            self.joint.pos_shaping_pre[env_index] = (
                torch.linalg.vector_norm(
                    self.joint.landmark.state.pos[env_index]
                    - self.pass_center[env_index]
                )
                * self.pos_shaping_factor
            )
            self.joint.pos_shaping_post[env_index] = (
                torch.linalg.vector_norm(
                    self.joint.landmark.state.pos[env_index]
                    - self.goal.state.pos[env_index],
                )
                * self.pos_shaping_factor
            )
            self.joint.rot_shaping_pre[env_index] = (
                get_line_angle_dist_0_360(
                    self.joint.landmark.state.rot[env_index].unsqueeze(-1),
                    self.middle_angle[env_index].unsqueeze(-1),
                )
                * self.rot_shaping_factor
            )
            # self.joint.rot_shaping_post[env_index] = (
            #     get_line_angle_dist_0_180(
            #         self.joint.landmark.state.rot[env_index],
            #         self.goal.state.rot[env_index],
            #     )
            #     * self.rot_shaping_factor
            # )

    def reward(self, agent: Agent):
        is_first = agent == self.world.agents[0]

        if is_first:
            self.rew = torch.zeros(
                self.world.batch_dim, device=self.world.device, dtype=torch.float32
            )
            self.pos_rew[:] = 0
            self.rot_rew[:] = 0
            self.collision_rew[:] = 0
            self.energy_rew[:] = 0

            joint_passed = self.joint.landmark.state.pos[:, Y] > 0
            self.all_passed = (
                torch.stack([a.state.pos[:, Y] for a in self.world.agents], dim=1)
                > self.passage_width / 2
            ).all(dim=1)

            # Pos shaping
            joint_dist_to_closest_pass = (
                torch.linalg.vector_norm(
                    self.joint.landmark.state.pos - self.pass_center, dim=1
                )
                * self.pos_shaping_factor
            )
            joint_shaping = joint_dist_to_closest_pass * self.pos_shaping_factor
            self.pos_rew[~joint_passed] += (self.joint.pos_shaping_pre - joint_shaping)[
                ~joint_passed
            ]
            self.joint.pos_shaping_pre = joint_shaping

            joint_dist_to_goal = torch.linalg.vector_norm(
                self.joint.landmark.state.pos - self.goal.state.pos,
                dim=1,
            )
            joint_shaping = joint_dist_to_goal * self.pos_shaping_factor
            self.pos_rew[joint_passed] += (self.joint.pos_shaping_post - joint_shaping)[
                joint_passed
            ]
            self.joint.pos_shaping_post = joint_shaping

            # Rot shaping
            joint_dist_to_90_rot = get_line_angle_dist_0_360(
                self.joint.landmark.state.rot, self.middle_angle
            )
            joint_shaping = joint_dist_to_90_rot * self.rot_shaping_factor
            self.rot_rew += self.joint.rot_shaping_pre - joint_shaping
            self.joint.rot_shaping_pre = joint_shaping

            # joint_dist_to_goal_rot = get_line_angle_dist_0_180(
            #     self.joint.landmark.state.rot, self.goal.state.rot
            # )
            # joint_shaping = joint_dist_to_goal_rot * self.rot_shaping_factor
            # self.rot_rew[rot_passed] += (self.joint.rot_shaping_post - joint_shaping)[
            #     rot_passed
            # ]
            # self.joint.rot_shaping_post = joint_shaping

            # Agent collisions
            if self.collision_reward != 0:
                for a in self.world.agents + ([self.mass] if self.asym_package else []):
                    for passage in self.passages:
                        if passage.collide:
                            self.collision_rew[
                                self.world.get_distance(a, passage)
                                <= self.min_collision_distance
                            ] += self.collision_reward
                    for wall in self.walls:
                        self.collision_rew[
                            self.world.get_distance(a, wall)
                            <= self.min_collision_distance
                        ] += self.collision_reward

            # Energy reward
            if self.energy_reward_coeff != 0:
                self.energy_expenditure = torch.stack(
                    [
                        torch.linalg.vector_norm(a.action.u, dim=-1)
                        / math.sqrt(
                            self.world.dim_p * ((a.u_range * a.u_multiplier) ** 2)
                        )
                        for a in self.world.agents
                    ],
                    dim=1,
                ).sum(-1)
                self.energy_rew = -self.energy_expenditure * self.energy_reward_coeff

            self.rew = (
                self.pos_rew + self.rot_rew + self.collision_rew + self.energy_rew
            )

        return self.rew

    def process_action(self, agent: Agent):
        vel_is_zero = torch.linalg.vector_norm(agent.action.u, dim=1) < 1e-3
        agent.controller.reset(vel_is_zero)
        agent.controller.process_force()

    def is_out_or_touching_perimeter(self, agent: Agent):
        is_out_or_touching_perimeter = torch.full(
            (self.world.batch_dim,), False, device=self.world.device
        )
        is_out_or_touching_perimeter += agent.state.pos[:, X] >= self.world.x_semidim
        is_out_or_touching_perimeter += agent.state.pos[:, X] <= -self.world.x_semidim
        is_out_or_touching_perimeter += agent.state.pos[:, Y] >= self.world.y_semidim
        is_out_or_touching_perimeter += agent.state.pos[:, Y] <= -self.world.y_semidim
        return is_out_or_touching_perimeter

    def observation(self, agent: Agent):
        if self.observe_joint_angle:
            joint_angle = self.joint.landmark.state.rot
            angle_noise = (
                torch.randn(
                    *joint_angle.shape, device=self.world.device, dtype=torch.float32
                )
                * self.joint_angle_obs_noise
                if self.joint_angle_obs_noise
                else 0.0
            )
            joint_angle += angle_noise

        observations = [
            agent.state.pos,
            agent.state.vel,
            agent.state.pos - self.goal.state.pos,
            agent.state.pos - self.big_passage_pos,
            agent.state.pos - self.small_passage_pos,
            angle_to_vector(self.goal.state.rot),
        ] + ([angle_to_vector(joint_angle)] if self.observe_joint_angle else [])

        if self.obs_noise > 0:
            for i, obs in enumerate(observations):
                noise = torch.zeros(*obs.shape, device=self.world.device,).uniform_(
                    -self.obs_noise,
                    self.obs_noise,
                )
                observations[i] = obs + noise
        return torch.cat(
            observations,
            dim=-1,
        )

    def done(self):
        return torch.all(
            (
                torch.linalg.vector_norm(
                    self.joint.landmark.state.pos - self.goal.state.pos, dim=1
                )
                <= 0.01
            )
            * (
                get_line_angle_dist_0_180(
                    self.joint.landmark.state.rot, self.goal.state.rot
                ).unsqueeze(-1)
                <= 0.01
            ),
            dim=1,
        )

    def info(self, agent: Agent) -> Dict[str, Tensor]:
        is_first = self.world.agents[0] == agent
        if is_first:
            just_passed = self.all_passed * (self.passed == 0)
            self.passed[just_passed] = 100
            self.info_stored = {
                "pos_rew": self.pos_rew,
                "rot_rew": self.rot_rew,
                "collision_rew": self.collision_rew,
                "energy_rew": self.energy_rew,
                "passed": just_passed.to(torch.int),
            }
        return self.info_stored

    def create_passage_map(self, world: World):
        # Add landmarks
        self.passages = []
        self.collide_passages = []
        self.non_collide_passages = []
        self.n_passages = 3

        def is_passage(i):
            return i < self.n_passages

        for i in range(self.n_boxes):
            passage = Landmark(
                name=f"passage {i}",
                collide=not is_passage(i),
                movable=False,
                shape=Box(length=self.passage_length, width=self.passage_width),
                color=Color.RED,
                collision_filter=lambda e: not isinstance(e.shape, Box),
            )
            if not passage.collide:
                self.non_collide_passages.append(passage)
            else:
                self.collide_passages.append(passage)
            self.passages.append(passage)
            world.add_landmark(passage)

    def spawn_passage_map(self, env_index):
        if self.fixed_passage:
            big_passage_start_index = torch.full(
                (self.world.batch_dim,) if env_index is None else (1,),
                5,
                device=self.world.device,
            )
            small_left_or_right = torch.full(
                (self.world.batch_dim,) if env_index is None else (1,),
                1,
                device=self.world.device,
            )
        else:
            big_passage_start_index = torch.randint(
                0,
                self.n_boxes - 1,
                (self.world.batch_dim,) if env_index is None else (1,),
                device=self.world.device,
            )
            small_left_or_right = torch.randint(
                0,
                2,
                (self.world.batch_dim,) if env_index is None else (1,),
                device=self.world.device,
            )

        small_left_or_right[big_passage_start_index > self.n_boxes - 1 - 4] = 0
        small_left_or_right[big_passage_start_index < 3] = 1
        small_left_or_right[small_left_or_right == 0] -= 3
        small_left_or_right[small_left_or_right == 1] += 3

        def is_passage(i):
            is_pass = big_passage_start_index == i
            is_pass += big_passage_start_index == i - 1
            is_pass += big_passage_start_index + small_left_or_right == i
            return is_pass

        def get_pos(i):
            pos = torch.tensor(
                [
                    -1 - self.agent_radius + self.passage_length / 2,
                    0.0,
                ],
                dtype=torch.float32,
                device=self.world.device,
            ).repeat(i.shape[0], 1)
            pos[:, X] += self.passage_length * i
            return pos

        for index, i in enumerate(
            [
                big_passage_start_index,
                big_passage_start_index + 1,
                big_passage_start_index + small_left_or_right,
            ]
        ):
            self.non_collide_passages[index].is_rendering[:] = False
            self.non_collide_passages[index].set_pos(get_pos(i), batch_index=env_index)

        big_passage_pos = (
            get_pos(big_passage_start_index) + get_pos(big_passage_start_index + 1)
        ) / 2
        small_passage_pos = get_pos(big_passage_start_index + small_left_or_right)
        pass_center = (big_passage_pos + small_passage_pos) / 2

        if env_index is None:
            self.small_left_or_right = small_left_or_right
            self.pass_center = pass_center
            self.big_passage_pos = big_passage_pos
            self.small_passage_pos = small_passage_pos
            self.middle_angle[small_left_or_right > 0] = torch.pi
            self.middle_angle[small_left_or_right < 0] = 0
        else:
            self.pass_center[env_index] = pass_center
            self.small_left_or_right[env_index] = small_left_or_right
            self.big_passage_pos[env_index] = big_passage_pos
            self.small_passage_pos[env_index] = small_passage_pos
            self.middle_angle[env_index] = (
                0 if small_left_or_right.item() < 0 else torch.pi
            )

        i = torch.zeros(
            (self.world.batch_dim,) if env_index is None else (1,),
            dtype=torch.int,
            device=self.world.device,
        )
        for passage in self.collide_passages:
            is_pass = is_passage(i)
            while is_pass.any():
                i[is_pass] += 1
                is_pass = is_passage(i)
            passage.set_pos(get_pos(i), batch_index=env_index)
            i += 1

    def spawn_walls(self, env_index):
        for i, wall in enumerate(self.walls):
            wall.set_pos(
                torch.tensor(
                    [
                        0.0
                        if i % 2
                        else (
                            self.world.x_semidim + self.agent_radius
                            if i == 0
                            else -self.world.x_semidim - self.agent_radius
                        ),
                        0.0
                        if not i % 2
                        else (
                            self.world.y_semidim + self.agent_radius
                            if i == 1
                            else -self.world.y_semidim - self.agent_radius
                        ),
                    ],
                    device=self.world.device,
                ),
                batch_index=env_index,
            )
            wall.set_rot(
                torch.tensor(
                    [torch.pi / 2 if not i % 2 else 0.0], device=self.world.device
                ),
                batch_index=env_index,
            )

    def extra_render(self, env_index: int = 0):
        from vmas.simulator import rendering

        geoms = []

        # Agent goal circles
        color = self.goal.color
        if isinstance(color, torch.Tensor) and len(color.shape) > 1:
            color = color[env_index]
        goal_agent_1 = rendering.make_circle(self.agent_radius)
        xform = rendering.Transform()
        goal_agent_1.add_attr(xform)
        xform.set_translation(
            self.goal.state.pos[env_index][X]
            - self.joint_length / 2 * math.cos(self.goal.state.rot[env_index]),
            self.goal.state.pos[env_index][Y]
            - self.joint_length / 2 * math.sin(self.goal.state.rot[env_index]),
        )

        goal_agent_1.set_color(*color)
        geoms.append(goal_agent_1)
        goal_agent_2 = rendering.make_circle(self.agent_radius)
        xform = rendering.Transform()
        goal_agent_2.add_attr(xform)
        xform.set_translation(
            self.goal.state.pos[env_index][X]
            + self.joint_length / 2 * math.cos(self.goal.state.rot[env_index]),
            self.goal.state.pos[env_index][Y]
            + self.joint_length / 2 * math.sin(self.goal.state.rot[env_index]),
        )
        goal_agent_2.set_color(*color)
        geoms.append(goal_agent_2)

        return geoms


if __name__ == "__main__":
    render_interactively(__file__, control_two_agents=True)



#  Copyright (c) 2022-2023.
#  ProrokLab (https://www.proroklab.org/)
#  All rights reserved.
from typing import Dict

import torch
from torch import Tensor

from vmas import render_interactively
from vmas.simulator.core import Agent, Landmark, Sphere, World
from vmas.simulator.joints import Joint
from vmas.simulator.scenario import BaseScenario
from vmas.simulator.utils import X, Color, JOINT_FORCE


class Scenario(BaseScenario):
    def make_world(self, batch_dim: int, device: torch.device, **kwargs):
        self.pos_shaping_factor = kwargs.get("pos_shaping_factor", 0)
        self.speed_shaping_factor = kwargs.get("speed_shaping_factor", 1)
        self.dist_shaping_factor = kwargs.get("dist_shaping_factor", 0)
        self.joints = kwargs.get("joints", True)

        self.n_agents = 2

        self.desired_speed = 1
        self.desired_radius = 0.5

        self.agent_spacing = 0.4
        self.agent_radius = 0.03
        self.ball_radius = 2 * self.agent_radius

        # Make world
        world = World(
            batch_dim,
            device,
            substeps=15 if self.joints else 5,
            joint_force=900 if self.joints else JOINT_FORCE,
            collision_force=1500 if self.joints else 400,
            drag=0,
        )
        # Add agents
        agent = Agent(name="agent 0", shape=Sphere(self.agent_radius), drag=0.25)
        world.add_agent(agent)
        agent = Agent(name="agent 1", shape=Sphere(self.agent_radius), drag=0.25)
        world.add_agent(agent)

        self.ball = Landmark(
            name="ball",
            shape=Sphere(radius=self.ball_radius),
            collide=True,
            movable=True,
            linear_friction=0.04,
        )
        world.add_landmark(self.ball)

        if self.joints:
            self.joints = []
            for i in range(self.n_agents):
                self.joints.append(
                    Joint(
                        world.agents[i],
                        self.ball,
                        anchor_a=(0, 0),
                        anchor_b=(0, 0),
                        dist=self.agent_spacing / 2,
                        rotate_a=True,
                        rotate_b=True,
                        collidable=False,
                        width=0,
                        mass=1,
                    )
                )
                world.add_joint(self.joints[i])

        self.pos_rew = torch.zeros(batch_dim, device=device, dtype=torch.float32)
        self.speed_rew = self.pos_rew.clone()
        self.dist_rew = self.pos_rew.clone()

        return world

    def reset_world_at(self, env_index: int = None):

        ball_pos = torch.zeros(
            (1, self.world.dim_p)
            if env_index is not None
            else (self.world.batch_dim, self.world.dim_p),
            device=self.world.device,
            dtype=torch.float32,
        ).uniform_(
            -self.desired_radius,
            self.desired_radius,
        )

        self.ball.set_pos(
            ball_pos,
            batch_index=env_index,
        )

        order = torch.randperm(self.n_agents).tolist()
        agents = [self.world.agents[i] for i in order]
        for i, agent in enumerate(agents):
            agent_pos = ball_pos.clone()
            agent_pos[:, X] += (self.agent_spacing / 2) * (-1 if i == 0 else 1)
            agent.set_pos(
                agent_pos,
                batch_index=env_index,
            )

        if env_index is None:
            self.pos_shaping = (
                torch.linalg.vector_norm(
                    self.ball.state.pos
                    - self.get_closest_point_circle(self.ball.state.pos),
                    dim=1,
                )
                ** 0.5
                * self.pos_shaping_factor
            )

            self.speed_shaping = (
                self.desired_speed
                - torch.linalg.vector_norm(self.ball.state.vel, dim=1)
            ).abs() * self.speed_shaping_factor

            self.dist_shaping = (
                torch.stack(
                    [
                        torch.linalg.vector_norm(
                            a.state.pos - self.ball.state.pos, dim=1
                        )
                        for a in self.world.agents
                    ],
                    dim=1,
                ).sum(dim=1)
                * self.dist_shaping_factor
            )
        else:
            self.pos_shaping = (
                torch.linalg.vector_norm(
                    self.ball.state.pos[env_index]
                    - self.get_closest_point_circle(self.ball.state.pos)[env_index]
                )
                ** 0.5
                * self.pos_shaping_factor
            )

            self.speed_shaping[env_index] = (
                self.desired_speed
                - torch.linalg.vector_norm(self.ball.state.vel[env_index])
            ).abs() * self.speed_shaping_factor

            self.dist_shaping[env_index] = (
                torch.stack(
                    [
                        torch.linalg.vector_norm(
                            a.state.pos[env_index] - self.ball.state.pos[env_index],
                        ).unsqueeze(-1)
                        for a in self.world.agents
                    ],
                    dim=1,
                ).sum(dim=1)
                * self.dist_shaping_factor
            )

    def reward(self, agent: Agent):

        pos_shaping = (
            torch.linalg.vector_norm(
                self.ball.state.pos
                - self.get_closest_point_circle(self.ball.state.pos),
                dim=1,
            )
            ** 0.5
            * self.pos_shaping_factor
        )
        self.pos_rew = self.pos_shaping - pos_shaping
        self.pos_shaping = pos_shaping

        speed = torch.linalg.vector_norm(self.ball.state.vel, dim=1)
        speed_shaping = (self.desired_speed - speed).abs() * self.speed_shaping_factor
        self.speed_rew = self.speed_shaping - speed_shaping
        self.speed_shaping = speed_shaping

        dist_shaping = (
            torch.stack(
                [
                    torch.linalg.vector_norm(a.state.pos - self.ball.state.pos, dim=1)
                    for a in self.world.agents
                ],
                dim=1,
            ).sum(dim=1)
            * self.dist_shaping_factor
        )
        self.dist_rew = self.dist_shaping - dist_shaping
        self.dist_shaping = dist_shaping

        return self.pos_rew + self.speed_rew + self.dist_rew

    def observation(self, agent: Agent):
        # get positions of all entities in this agent's reference frame
        return torch.cat(
            [
                agent.state.pos,
                agent.state.vel,
                agent.state.pos - self.ball.state.pos,
                agent.state.pos,
            ],
            dim=-1,
        )

    def info(self, agent: Agent) -> Dict[str, Tensor]:
        return {
            "pos_rew": self.pos_rew,
            "speed_rew": self.speed_rew,
            "dist_rew": self.dist_rew,
        }

    def get_closest_point_circle(self, pos: Tensor):
        pos_norm = torch.linalg.vector_norm(pos, dim=1)
        agent_pos_normalized = pos / pos_norm.unsqueeze(-1)

        agent_pos_normalized *= self.desired_radius

        return torch.nan_to_num(agent_pos_normalized)

    def extra_render(self, env_index: int = 0):
        from vmas.simulator import rendering

        geoms = []

        # Trajectory goal circle
        color = Color.BLACK.value
        circle = rendering.make_circle(self.desired_radius, filled=False)
        xform = rendering.Transform()
        circle.add_attr(xform)
        xform.set_translation(0, 0)
        circle.set_color(*color)
        geoms.append(circle)

        return geoms


if __name__ == "__main__":
    render_interactively(__file__, control_two_agents=True)



#  Copyright (c) 2022.
#  ProrokLab (https://www.proroklab.org/)
#  All rights reserved.

import torch
from vmas import render_interactively

from vmas.simulator.core import World, Agent, Landmark, Sphere
from vmas.simulator.scenario import BaseScenario
from vmas.simulator.utils import Color


class Scenario(BaseScenario):
    def make_world(self, batch_dim: int, device: torch.device, **kwargs):
        n_agents = kwargs.get("n_agents", 3)
        n_adversaries = kwargs.get("n_adversaries", 1)
        assert n_agents > n_adversaries

        world = World(
            batch_dim=batch_dim,
            device=device,
        )
        num_adversaries = n_adversaries
        num_landmarks = n_agents - 1

        # Add agents
        for i in range(n_agents):
            adversary = True if i < num_adversaries else False
            agent = Agent(
                name=f"agent {i}",
                collide=False,
                shape=Sphere(radius=0.15),
                color=Color.RED if adversary else Color.BLUE,
                adversary=adversary,
            )
            world.add_agent(agent)
        # Add landmarks
        for i in range(num_landmarks):
            landmark = Landmark(
                name=f"landmark {i}",
                collide=False,
                shape=Sphere(radius=0.08),
                color=Color.BLACK,
            )
            world.add_landmark(landmark)

        return world

    def reset_world_at(self, env_index: int = None):
        # set goal landmark
        if env_index is None:
            goal = self.world.landmarks[
                torch.randint(0, len(self.world.landmarks), (1,)).item()
            ]
            goal.color = Color.GREEN
            for agent in self.world.agents:
                agent.goal = goal

        for agent in self.world.agents:
            agent.set_pos(
                2
                * torch.rand(
                    self.world.dim_p, device=self.world.device, dtype=torch.float32
                )
                - 1,
                batch_index=env_index,
            )

        for landmark in self.world.landmarks:
            landmark.set_pos(
                2
                * torch.rand(
                    self.world.dim_p, device=self.world.device, dtype=torch.float32
                )
                - 1,
                batch_index=env_index,
            )

    # return all agents that are not adversaries
    def good_agents(self):
        return [agent for agent in self.world.agents if not agent.adversary]

    # return all adversarial agents
    def adversaries(self):
        return [agent for agent in self.world.agents if agent.adversary]

    def reward(self, agent: Agent):
        # Agents are rewarded based on minimum agent distance to each landmark
        return (
            self.adversary_reward(agent)
            if agent.adversary
            else self.agent_reward(agent)
        )

    def agent_reward(self, agent: Agent):
        # Rewarded based on how close any good agent is to the goal landmark, and how far the adversary is from it
        shaped_reward = True
        shaped_adv_reward = True

        # Calculate negative reward for adversary
        adversary_agents = self.adversaries()
        if shaped_adv_reward:  # distance-based adversary reward
            adv_rew = torch.sum(
                torch.stack(
                    [
                        torch.sqrt(
                            torch.sum(
                                torch.square(a.state.pos - a.goal.state.pos), dim=-1
                            )
                        )
                        for a in adversary_agents
                    ],
                    dim=1,
                ),
                dim=-1,
            )
        else:  # proximity-based adversary reward (binary)
            adv_rew = torch.zeros(
                self.world.batch_dim, device=self.world.device, dtype=torch.float32
            )
            for a in adversary_agents:
                is_too_close = (
                    torch.sqrt(
                        torch.sum(torch.square(a.state.pos - a.goal.state.pos), dim=-1)
                    )
                    < 2 * a.goal.size
                )
                adv_rew[is_too_close] -= 5

        # Calculate positive reward for agents
        good_agents = self.good_agents()
        if shaped_reward:  # distance-based agent reward
            pos_rew = -torch.min(
                torch.stack(
                    [
                        torch.sqrt(
                            torch.sum(
                                torch.square(a.state.pos - a.goal.state.pos), dim=-1
                            )
                        )
                        for a in good_agents
                    ],
                    dim=1,
                ),
                dim=-1,
            )[0]

        else:  # proximity-based agent reward (binary)
            pos_rew = torch.zeros(
                self.world.batch_dim, device=self.world.device, dtype=torch.float32
            )
            is_close_enough = (
                torch.min(
                    torch.stack(
                        [
                            torch.sqrt(
                                torch.sum(
                                    torch.square(a.state.pos - a.goal.state.pos), dim=-1
                                )
                            )
                            for a in good_agents
                        ],
                        dim=1,
                    ),
                    dim=-1,
                )
                < 2 * agent.goal.size
            )
            pos_rew[is_close_enough] += 5
            pos_rew -= torch.min(
                torch.stack(
                    [
                        torch.sqrt(
                            torch.sum(
                                torch.square(a.state.pos - a.goal.state.pos), dim=-1
                            )
                        )
                        for a in good_agents
                    ],
                    dim=1,
                ),
                dim=-1,
            )
        return pos_rew + adv_rew

    def adversary_reward(self, agent: Agent):
        # Rewarded based on proximity to the goal landmark
        shaped_reward = True
        if shaped_reward:  # distance-based reward
            return -torch.sqrt(
                torch.sum(torch.square(agent.state.pos - agent.goal.state.pos), dim=-1)
            )

        else:  # proximity-based reward (binary)
            adv_rew = torch.zeros(
                self.world.batch_dim, device=self.world.device, dtype=torch.float32
            )
            close_enough = (
                torch.sqrt(
                    torch.sum(
                        torch.square(agent.state.pos - agent.goal.state.pos), dim=-1
                    )
                )
                < 2 * agent.goal.size
            )
            adv_rew[close_enough] += 5
            return adv_rew

    def observation(self, agent: Agent):
        # get positions of all entities in this agent's reference frame
        entity_pos = []
        for entity in self.world.landmarks:
            entity_pos.append(entity.state.pos - agent.state.pos)
        # position of all other agents
        other_pos = []
        for other in self.world.agents:
            if other is agent:
                continue
            other_pos.append(other.state.pos - agent.state.pos)

        if not agent.adversary:
            return torch.cat(
                [agent.goal.state.pos - agent.state.pos, *entity_pos, *other_pos],
                dim=-1,
            )
        else:
            return torch.cat([*entity_pos, *other_pos], dim=-1)


if __name__ == "__main__":
    render_interactively(__file__, control_two_agents=True)



#  Copyright (c) 2022-2023.
#  ProrokLab (https://www.proroklab.org/)
#  All rights reserved.
import math
from typing import Dict

import torch
from torch import Tensor

from vmas import render_interactively
from vmas.simulator.core import Agent, World, Landmark, Sphere
from vmas.simulator.scenario import BaseScenario
from vmas.simulator.utils import Color, TorchUtils
from vmas.simulator.velocity_controller import VelocityController


class Scenario(BaseScenario):
    def make_world(self, batch_dim: int, device: torch.device, **kwargs):
        self.u_range = kwargs.get("u_range", 1)
        self.a_range = kwargs.get("a_range", 1)
        self.obs_noise = kwargs.get("obs_noise", 0.0)
        self.dt_delay = kwargs.get("dt_delay", 0)
        self.min_input_norm = kwargs.get("min_input_norm", 0.08)
        self.linear_friction = kwargs.get("linear_friction", 0.1)

        self.pos_shaping_factor = kwargs.get("pos_shaping_factor", 1.0)
        self.time_rew_coeff = kwargs.get("time_rew_coeff", -0.01)
        self.energy_reward_coeff = kwargs.get("energy_rew_coeff", 0.0)

        self.viewer_size = (1600, 700)
        self.viewer_zoom = 2

        self.plot_grid = True
        self.agent_radius = 0.16

        self.lab_length = 6
        self.lab_width = 3

        controller_params = [2, 6, 0.002]

        self.f_range = self.a_range + self.linear_friction

        # Make world
        world = World(
            batch_dim,
            device,
            drag=0,
            dt=0.05,
            substeps=5,
        )

        null_action = torch.zeros(world.batch_dim, world.dim_p, device=world.device)
        self.input_queue = [null_action.clone() for _ in range(self.dt_delay)]
        # control delayed by n dts

        # Goal
        self.goal = Landmark(
            "goal",
            collide=False,
            movable=False,
            shape=Sphere(radius=0.06),
        )

        world.add_landmark(self.goal)
        # Add agents
        agent = Agent(
            name=f"agent 0",
            collide=True,
            color=Color.GREEN,
            render_action=True,
            linear_friction=self.linear_friction,
            shape=Sphere(radius=self.agent_radius),
            f_range=self.f_range,
            u_range=self.u_range,
        )
        agent.controller = VelocityController(
            agent, world, controller_params, "standard"
        )
        agent.goal = self.goal
        agent.energy_rew = torch.zeros(batch_dim, device=device)
        world.add_agent(agent)

        self.pos_rew = torch.zeros(batch_dim, device=device)
        self.time_rew = self.pos_rew.clone()

        return world

    def reset_world_at(self, env_index: int = None):
        for agent in self.world.agents:
            agent.controller.reset(env_index)
            agent.set_pos(
                torch.cat(
                    [
                        torch.zeros(
                            (1, 1)
                            if env_index is not None
                            else (self.world.batch_dim, 1),
                            device=self.world.device,
                            dtype=torch.float32,
                        ).uniform_(
                            -self.lab_length / 2,
                            self.lab_length / 2,
                        ),
                        torch.zeros(
                            (1, 1)
                            if env_index is not None
                            else (self.world.batch_dim, 1),
                            device=self.world.device,
                            dtype=torch.float32,
                        ).uniform_(
                            -self.lab_width / 2,
                            self.lab_width / 2,
                        ),
                    ],
                    dim=1,
                ),
                batch_index=env_index,
            )

        for landmark in self.world.landmarks:
            # Random pos between -1 and 1
            landmark.set_pos(
                torch.cat(
                    [
                        torch.zeros(
                            (1, 1)
                            if env_index is not None
                            else (self.world.batch_dim, 1),
                            device=self.world.device,
                            dtype=torch.float32,
                        ).uniform_(
                            -self.lab_length / 2,
                            self.lab_length / 2,
                        ),
                        torch.zeros(
                            (1, 1)
                            if env_index is not None
                            else (self.world.batch_dim, 1),
                            device=self.world.device,
                            dtype=torch.float32,
                        ).uniform_(
                            -self.lab_width / 2,
                            self.lab_width / 2,
                        ),
                    ],
                    dim=1,
                ),
                batch_index=env_index,
            )
            if env_index is None:
                landmark.pos_shaping = (
                    torch.stack(
                        [
                            torch.linalg.vector_norm(
                                landmark.state.pos - a.state.pos, dim=1
                            )
                            for a in self.world.agents
                        ],
                        dim=1,
                    ).min(dim=1)[0]
                    * self.pos_shaping_factor
                )
            else:
                landmark.pos_shaping[env_index] = (
                    torch.stack(
                        [
                            torch.linalg.vector_norm(
                                landmark.state.pos[env_index] - a.state.pos[env_index],
                            ).unsqueeze(-1)
                            for a in self.world.agents
                        ],
                        dim=1,
                    ).min(dim=1)[0]
                    * self.pos_shaping_factor
                )

    def process_action(self, agent: Agent):
        # Use queue for delay
        self.input_queue.append(agent.action.u.clone())
        agent.action.u = self.input_queue.pop(0)

        # Clamp square to circle
        agent.action.u = TorchUtils.clamp_with_norm(agent.action.u, self.u_range)

        # Zero small input
        action_norm = torch.linalg.vector_norm(agent.action.u, dim=1)
        agent.action.u[action_norm < self.min_input_norm] = 0

        agent.vel_action = agent.action.u.clone()
        agent.controller.process_force()

    def reward(self, agent: Agent):
        is_first = agent == self.world.agents[0]

        if is_first:
            self.pos_rew[:] = 0
            self.time_rew[:] = 0

            # Pos shaping
            goal_dist = torch.stack(
                [
                    torch.linalg.vector_norm(self.goal.state.pos - a.state.pos, dim=1)
                    for a in self.world.agents
                ],
                dim=1,
            ).min(dim=1)[0]

            self.goal_reached = goal_dist < self.goal.shape.radius
            pos_shaping = goal_dist * self.pos_shaping_factor
            self.pos_rew[~self.goal_reached] = (self.goal.pos_shaping - pos_shaping)[
                ~self.goal_reached
            ]
            self.goal.pos_shaping = pos_shaping

            self.time_rew[~self.goal_reached] += self.time_rew_coeff

        # Energy reward
        agent.energy_expenditure = torch.stack(
            [
                torch.linalg.vector_norm(a.action.u, dim=-1)
                / math.sqrt(self.world.dim_p * (a.f_range**2))
                for a in self.world.agents
            ],
            dim=1,
        ).sum(-1)
        agent.energy_rew = -agent.energy_expenditure * self.energy_reward_coeff

        return self.pos_rew + agent.energy_rew + self.time_rew

    def observation(self, agent: Agent):
        observations = [
            agent.state.pos,
            agent.state.vel,
            agent.state.pos - self.goal.state.pos,
        ]

        if self.obs_noise > 0:
            for i, obs in enumerate(observations):
                noise = torch.zeros(*obs.shape, device=self.world.device,).uniform_(
                    -self.obs_noise,
                    self.obs_noise,
                )
                observations[i] = obs + noise
        return torch.cat(
            observations,
            dim=-1,
        )

    def info(self, agent: Agent) -> Dict[str, Tensor]:
        return {
            "pos_rew": self.pos_rew,
            "energy_rew": agent.energy_rew,
            "time_rew": self.time_rew,
        }


if __name__ == "__main__":
    render_interactively(__file__, control_two_agents=False)



#  Copyright (c) 2022-2023.
#  ProrokLab (https://www.proroklab.org/)
#  All rights reserved.
import math
from typing import Dict

import numpy as np
import torch
from torch import Tensor

from vmas import render_interactively
from vmas.simulator.core import Agent, World
from vmas.simulator.scenario import BaseScenario
from vmas.simulator.utils import Color, Y


class Scenario(BaseScenario):
    def make_world(self, batch_dim: int, device: torch.device, **kwargs):
        self.green_mass = kwargs.get("green_mass", 4)
        self.blue_mass = kwargs.get("blue_mass", 2)
        self.mass_noise = kwargs.get("mass_noise", 1)
        self.plot_grid = True

        # Make world
        world = World(batch_dim, device)
        # Add agents
        self.green_agent = Agent(
            name=f"agent 0",
            collide=False,
            color=Color.GREEN,
            render_action=True,
            mass=self.green_mass,
            f_range=1,
        )
        world.add_agent(self.green_agent)
        self.blue_agent = Agent(
            name=f"agent 1", collide=False, render_action=True, f_range=1
        )
        world.add_agent(self.blue_agent)

        self.max_speed = torch.zeros(batch_dim, device=device)
        self.energy_expenditure = self.max_speed.clone()

        return world

    def reset_world_at(self, env_index: int = None):
        # Temp
        self.blue_agent.mass = self.blue_mass + np.random.uniform(
            -self.mass_noise, self.mass_noise
        )
        self.green_agent.mass = self.green_mass + np.random.uniform(
            -self.mass_noise, self.mass_noise
        )

        for agent in self.world.agents:
            agent.set_pos(
                torch.zeros(
                    (1, self.world.dim_p)
                    if env_index is not None
                    else (self.world.batch_dim, self.world.dim_p),
                    device=self.world.device,
                    dtype=torch.float32,
                ).uniform_(-1, 1),
                batch_index=env_index,
            )

    def process_action(self, agent: Agent):
        agent.action.u[:, Y] = 0

    def reward(self, agent: Agent):
        is_first = agent == self.world.agents[0]

        if is_first:
            self.max_speed = torch.stack(
                [
                    torch.linalg.vector_norm(a.state.vel, dim=1)
                    for a in self.world.agents
                ],
                dim=1,
            ).max(dim=1)[0]

            self.energy_expenditure = (
                -torch.stack(
                    [
                        torch.linalg.vector_norm(a.action.u, dim=-1)
                        / math.sqrt(self.world.dim_p * (a.f_range**2))
                        for a in self.world.agents
                    ],
                    dim=1,
                ).sum(-1)
                * 0.17
            )

            # print(self.max_speed)
            # print(self.energy_expenditure)
            # self.energy_rew_1 = (self.world.agents[0].action.u[:, X] - 0).abs()
            # self.energy_rew_1 += (self.world.agents[0].action.u[:, Y] - 0).abs()
            #
            # self.energy_rew_2 = (self.world.agents[1].action.u[:, X] - 0).abs()
            # self.energy_rew_2 += (self.world.agents[1].action.u[:, Y] - 0).abs()

        return self.max_speed + self.energy_expenditure

    def observation(self, agent: Agent):
        return torch.cat(
            [agent.state.pos, agent.state.vel],
            dim=-1,
        )

    def info(self, agent: Agent) -> Dict[str, Tensor]:
        return {
            "max_speed": self.max_speed,
            "energy_expenditure": self.energy_expenditure,
        }


if __name__ == "__main__":
    render_interactively(__file__, control_two_agents=True)



#  Copyright (c) 2022-2023.
#  ProrokLab (https://www.proroklab.org/)
#  All rights reserved.
import math
import typing
from typing import Dict, List

import torch
from torch import Tensor

from vmas import render_interactively
from vmas.simulator.core import Agent, Landmark, Sphere, World
from vmas.simulator.joints import Joint
from vmas.simulator.scenario import BaseScenario
from vmas.simulator.utils import Color

if typing.TYPE_CHECKING:
    from vmas.simulator.rendering import Geom


def get_line_angle_0_90(rot: Tensor):
    angle = torch.abs(rot) % torch.pi
    other_angle = torch.pi - angle
    return torch.minimum(angle, other_angle)


def get_line_angle_0_180(rot):
    angle = rot % torch.pi
    return angle


def get_line_angle_dist_0_180(angle, goal):
    angle = get_line_angle_0_180(angle)
    goal = get_line_angle_0_180(goal)
    return torch.minimum(
        (angle - goal).abs(),
        torch.minimum(
            (angle - (goal - torch.pi)).abs(), ((angle - torch.pi) - goal).abs()
        ),
    ).squeeze(-1)


def angle_to_vector(angle):
    return torch.cat([torch.cos(angle), torch.sin(angle)], dim=1)


class Scenario(BaseScenario):
    def make_world(self, batch_dim: int, device: torch.device, **kwargs):
        self.joint_length = kwargs.get("joint_length", 0.5)
        self.random_start_angle = kwargs.get("random_start_angle", False)
        self.observe_joint_angle = kwargs.get("observe_joint_angle", False)
        self.joint_angle_obs_noise = kwargs.get("joint_angle_obs_noise", 0.0)
        self.asym_package = kwargs.get("asym_package", True)
        self.mass_ratio = kwargs.get("mass_ratio", 5)
        self.mass_position = kwargs.get("mass_position", 0.75)
        self.max_speed_1 = kwargs.get("max_speed_1", None)  # 0.1
        self.obs_noise = kwargs.get("obs_noise", 0.2)

        # Reward
        self.rot_shaping_factor = kwargs.get("rot_shaping_factor", 1)
        self.energy_reward_coeff = kwargs.get("energy_reward_coeff", 0.08)

        # Make world
        world = World(
            batch_dim,
            device,
            substeps=7 if not self.asym_package else 10,
            joint_force=900 if self.asym_package else 400,
            drag=0.25 if not self.asym_package else 0.15,
        )

        if not self.observe_joint_angle:
            assert self.joint_angle_obs_noise == 0

        self.goal_angle = torch.pi / 2

        self.n_agents = 2

        self.agent_radius = 0.03333
        self.mass_radius = self.agent_radius * (2 / 3)

        # Add agents
        agent = Agent(
            name=f"agent 0",
            shape=Sphere(self.agent_radius),
            u_multiplier=0.8,
            obs_noise=self.obs_noise,
            render_action=True,
        )
        world.add_agent(agent)
        agent = Agent(
            name=f"agent 1",
            shape=Sphere(self.agent_radius),
            u_multiplier=0.8,
            mass=1 if self.asym_package else self.mass_ratio,
            max_speed=self.max_speed_1,
            obs_noise=self.obs_noise,
            render_action=True,
        )
        world.add_agent(agent)

        self.joint = Joint(
            world.agents[0],
            world.agents[1],
            anchor_a=(0, 0),
            anchor_b=(0, 0),
            dist=self.joint_length,
            rotate_a=True,
            rotate_b=True,
            collidable=False,
            width=0,
            mass=1,
        )
        world.add_joint(self.joint)

        if self.asym_package:

            def mass_collision_filter(e):
                return not isinstance(e.shape, Sphere)

            self.mass = Landmark(
                name="mass",
                shape=Sphere(radius=self.mass_radius),
                collide=False,
                movable=True,
                color=Color.BLACK,
                mass=self.mass_ratio,
                collision_filter=mass_collision_filter,
            )

            world.add_landmark(self.mass)

            joint = Joint(
                self.mass,
                self.joint.landmark,
                anchor_a=(0, 0),
                anchor_b=(self.mass_position, 0),
                dist=0,
                rotate_a=True,
                rotate_b=True,
            )
            world.add_joint(joint)

        self.rot_rew = torch.zeros(batch_dim, device=device)
        self.energy_rew = self.rot_rew.clone()

        return world

    def reset_world_at(self, env_index: int = None):
        start_angle = torch.zeros(
            (1, 1) if env_index is not None else (self.world.batch_dim, 1),
            device=self.world.device,
            dtype=torch.float32,
        ).uniform_(
            -torch.pi / 2 if self.random_start_angle else 0,
            torch.pi / 2 if self.random_start_angle else 0,
        )

        start_delta_x = (self.joint_length / 2) * torch.cos(start_angle)
        start_delta_x_abs = start_delta_x.abs()
        min_x_start = 0
        max_x_start = 0
        start_delta_y = (self.joint_length / 2) * torch.sin(start_angle)
        start_delta_y_abs = start_delta_y.abs()
        min_y_start = 0
        max_y_start = 0

        joint_pos = torch.cat(
            [
                (min_x_start - max_x_start)
                * torch.rand(
                    (1, 1) if env_index is not None else (self.world.batch_dim, 1),
                    device=self.world.device,
                    dtype=torch.float32,
                )
                + max_x_start,
                (min_y_start - max_y_start)
                * torch.rand(
                    (1, 1) if env_index is not None else (self.world.batch_dim, 1),
                    device=self.world.device,
                    dtype=torch.float32,
                )
                + max_y_start,
            ],
            dim=1,
        )

        order = torch.randperm(self.n_agents).tolist()
        agents = [self.world.agents[i] for i in order]
        for i, agent in enumerate(agents):
            if i == 0:
                agent.set_pos(
                    joint_pos - torch.cat([start_delta_x, start_delta_y], dim=1),
                    batch_index=env_index,
                )
            else:
                agent.set_pos(
                    joint_pos + torch.cat([start_delta_x, start_delta_y], dim=1),
                    batch_index=env_index,
                )

        if self.asym_package:
            self.mass.set_pos(
                joint_pos
                + self.mass_position
                * torch.cat([start_delta_x, start_delta_y], dim=1)
                * (1 if agents[0] == self.world.agents[0] else -1),
                batch_index=env_index,
            )

        if env_index is None:

            self.joint.rot_shaping_pre = (
                get_line_angle_dist_0_180(
                    self.joint.landmark.state.rot, self.goal_angle
                )
                * self.rot_shaping_factor
            )

        else:

            self.joint.rot_shaping_pre[env_index] = (
                get_line_angle_dist_0_180(
                    self.joint.landmark.state.rot[env_index], self.goal_angle
                )
                * self.rot_shaping_factor
            )

    def reward(self, agent: Agent):
        is_first = agent == self.world.agents[0]

        if is_first:
            self.rot_rew[:] = 0

            # Rot shaping
            joint_dist_to_90_rot = get_line_angle_dist_0_180(
                self.joint.landmark.state.rot, self.goal_angle
            )
            joint_shaping = joint_dist_to_90_rot * self.rot_shaping_factor
            self.rot_rew += self.joint.rot_shaping_pre - joint_shaping
            self.joint.rot_shaping_pre = joint_shaping

            # Energy reward
            self.energy_expenditure = torch.stack(
                [
                    torch.linalg.vector_norm(a.action.u, dim=-1)
                    / math.sqrt(self.world.dim_p * ((a.u_range * a.u_multiplier) ** 2))
                    for a in self.world.agents
                ],
                dim=1,
            ).sum(-1)
            self.energy_rew = -self.energy_expenditure * self.energy_reward_coeff

            self.rew = self.rot_rew + self.energy_rew

        return self.rew

    def observation(self, agent: Agent):
        if self.observe_joint_angle:
            joint_angle = self.joint.landmark.state.rot
            angle_noise = (
                torch.randn(
                    *joint_angle.shape, device=self.world.device, dtype=torch.float32
                )
                * self.joint_angle_obs_noise
                if self.joint_angle_obs_noise
                else 0.0
            )
            joint_angle += angle_noise

        observations = [
            agent.state.pos,
            agent.state.vel,
        ] + ([angle_to_vector(joint_angle)] if self.observe_joint_angle else [])

        for i, obs in enumerate(observations):
            noise = torch.zeros(*obs.shape, device=self.world.device,).uniform_(
                -self.obs_noise,
                self.obs_noise,
            )
            observations[i] = obs.clone() + noise
        return torch.cat(
            observations,
            dim=-1,
        )

    def done(self):
        return torch.all(
            (
                get_line_angle_dist_0_180(
                    self.joint.landmark.state.rot, self.goal_angle
                ).unsqueeze(-1)
                <= 0.01
            ),
            dim=1,
        )

    def info(self, agent: Agent) -> Dict[str, Tensor]:
        return {
            "rot_rew": self.rot_rew,
            "energy_rew": self.energy_rew,
        }

    def extra_render(self, env_index: int = 0) -> "List[Geom]":
        from vmas.simulator import rendering

        geoms = []

        color = Color.GREEN.value
        origin = rendering.make_circle(0.01)
        xform = rendering.Transform()
        origin.add_attr(xform)
        xform.set_translation(0, 0)
        origin.set_color(*color)
        geoms.append(origin)

        return geoms


if __name__ == "__main__":
    render_interactively(__file__, control_two_agents=True)



#  Copyright (c) 2022.
#  ProrokLab (https://www.proroklab.org/)
#  All rights reserved.
import time

import numpy as np
import torch

from vmas import make_env, Wrapper
from vmas.simulator.utils import save_video


def use_vmas_env(render: bool = False, save_render: bool = False):
    assert not (save_render and not render), "To save the video you have to render it"

    scenario_name = "waterfall"

    # Scenario specific variables
    n_agents = 4

    num_envs = 32
    continuous_actions = False
    device = "cpu"  # or cuda or any other torch device
    wrapper = None
    n_steps = 100

    simple_2d_action = (
        [0, 0.5] if continuous_actions else [3]
    )  # Sample action tell each agent to go down

    env = make_env(
        scenario_name=scenario_name,
        num_envs=num_envs,
        device=device,
        continuous_actions=continuous_actions,
        wrapper=wrapper,
        # Environment specific variables
        n_agents=n_agents,
    )

    frame_list = []  # For creating a gif
    init_time = time.time()
    step = 0
    for s in range(n_steps):
        actions = []
        step += 1
        print(f"Step {step}")
        if wrapper is Wrapper.RLLIB:  # Rllib interface
            for i in range(num_envs):
                actions_per_env = []
                for j in range(n_agents):
                    actions_per_env.append(np.array(simple_2d_action))
                actions.append(actions_per_env)
            obs, rews, dones, info = env.vector_step(actions)
            if render:
                frame_list.append(
                    env.try_render_at(
                        mode="rgb_array",
                        agent_index_focus=None,
                        visualize_when_rgb=True,
                    )  # Can give the camera an agent index to focus on
                )
        elif wrapper is None:  # Same as before, with faster VMAS interface
            for i in range(n_agents):
                actions.append(
                    torch.tensor(
                        simple_2d_action,
                        device=device,
                    ).repeat(num_envs, 1)
                )
            obs, rews, dones, info = env.step(actions)
            if render:
                frame_list.append(
                    env.render(
                        mode="rgb_array",
                        agent_index_focus=None,
                        visualize_when_rgb=True,
                    )
                )  # Can give the camera an agent index to focus on

    if render and save_render:
        if wrapper is not None:
            env = env.env
        save_video(scenario_name, frame_list, fps=1 / env.scenario.world.dt)

    total_time = time.time() - init_time
    print(
        f"It took: {total_time}s for {n_steps} steps of {num_envs} parallel environments on device {device}"
        f" for {wrapper.name + ' wrapped ' if wrapper is not None else ''}simulator"
    )


if __name__ == "__main__":
    use_vmas_env(render=True, save_render=False)



"""
Scenario:
1 speaker, 2 listeners (one of which is an adversary). Good agents rewarded for proximity to goal, and distance from
adversary to goal. Adversary is rewarded for its distance to the goal.
"""

#  Copyright (c) 2022.
#  ProrokLab (https://www.proroklab.org/)
#  All rights reserved.

import torch

from vmas.simulator.core import World, Agent
from vmas.simulator.scenario import BaseScenario
from vmas.simulator.utils import Color


class Scenario(BaseScenario):
    def make_world(self, batch_dim: int, device: torch.device, **kwargs):
        dim_c = kwargs.get("dim_c", 4)
        assert dim_c > 0

        world = World(
            batch_dim=batch_dim,
            device=device,
            dim_c=dim_c,
        )
        # set any world properties first
        num_agents = 3
        num_adversaries = 1

        # Add agents
        for i in range(num_agents):
            adversary = True if i < num_adversaries else False
            speaker = True if i == 2 else False
            agent = Agent(
                name=("eve" if adversary else ("alice" if speaker else "bob")),
                collide=False,
                movable=False,
                color=Color.RED
                if adversary
                else (Color.GREEN if speaker else Color.BLUE),
                adversary=adversary,
                silent=False,
            )
            agent.speaker = speaker
            world.add_agent(agent)

        return world

    def reset_world_at(self, env_index: int = None):
        key = torch.randint(
            0, 2, (self.world.batch_dim, self.world.dim_c), device=self.world.device
        )
        secret = torch.randint(
            0, 2, (self.world.batch_dim, self.world.dim_c), device=self.world.device
        )

        if env_index is None:
            for agent in self.world.agents:
                agent.key = None if not agent.speaker else key
                agent.secret = secret
                agent.set_pos(
                    2
                    * torch.rand(
                        self.world.dim_p, device=self.world.device, dtype=torch.float32
                    )
                    - 1,
                    batch_index=env_index,
                )
        else:
            for agent in self.world.agents:
                if agent.speaker:
                    agent.key[env_index] = key[env_index]
                agent.secret[env_index] = secret[env_index]

    # return all agents that are not adversaries
    def good_listeners(self):
        return [
            agent
            for agent in self.world.agents
            if not agent.adversary and not agent.speaker
        ]

    # return all agents that are not adversaries
    def good_agents(self):
        return [agent for agent in self.world.agents if not agent.adversary]

    # return all adversarial agents
    def adversaries(self):
        return [agent for agent in self.world.agents if agent.adversary]

    def reward(self, agent: Agent):
        return (
            self.adversary_reward(agent)
            if agent.adversary
            else self.agent_reward(agent)
        )

    def agent_reward(self, agent: Agent):
        # Agents rewarded if Bob can reconstruct message, but adversary (Eve) cannot
        good_listeners = self.good_listeners()
        adversaries = self.adversaries()
        good_rew = torch.zeros(
            self.world.batch_dim, device=self.world.device, dtype=torch.float32
        )
        adv_rew = torch.zeros(
            self.world.batch_dim, device=self.world.device, dtype=torch.float32
        )
        for a in good_listeners:
            zero_comms = torch.all(
                a.state.c
                == torch.zeros(
                    self.world.batch_dim,
                    self.world.dim_c,
                    device=self.world.device,
                    dtype=torch.float32,
                ),
                dim=-1,
            )
            good_rew[~zero_comms] -= torch.sum(
                torch.square(a.state.c - agent.secret), dim=-1
            )[~zero_comms]
        for a in adversaries:
            zero_comms = torch.all(
                a.state.c
                == torch.zeros(
                    self.world.batch_dim,
                    self.world.dim_c,
                    device=self.world.device,
                    dtype=torch.float32,
                ),
                dim=-1,
            )
            adv_rew[~zero_comms] += torch.sum(
                torch.square(a.state.c - agent.secret), dim=-1
            )[~zero_comms]
        return adv_rew + good_rew

    def adversary_reward(self, agent: Agent):
        # Adversary (Eve) is rewarded if it can reconstruct original goal
        rew = torch.zeros(
            self.world.batch_dim, device=self.world.device, dtype=torch.float32
        )
        zero_comms = torch.all(
            agent.state.c
            == torch.zeros(
                self.world.batch_dim,
                self.world.dim_c,
                device=self.world.device,
                dtype=torch.float32,
            ),
            dim=-1,
        )
        rew[~zero_comms] -= torch.sum(
            torch.square(agent.state.c - agent.secret), dim=-1
        )[~zero_comms]
        return rew

    def observation(self, agent: Agent):
        # communication of all other agents
        comm = []
        for other in self.world.agents:
            if other is agent or not other.speaker:
                continue
            comm.append(other.state.c)

        key = self.world.agents[2].key
        secret = self.world.agents[0].secret

        # speaker
        if agent.speaker:
            return torch.cat(
                [
                    secret,
                    key,
                ],
                dim=-1,
            )
        # listener
        if not agent.speaker and not agent.adversary:
            return torch.cat([key, *comm], dim=-1)
        # adv
        if not agent.speaker and agent.adversary:
            return torch.cat([*comm], dim=-1)



#  Copyright (c) 2022-2023.
#  ProrokLab (https://www.proroklab.org/)
#  All rights reserved.

from typing import Optional

from vmas import scenarios
from vmas.simulator.environment import Environment
from vmas.simulator.environment import Wrapper


def make_env(
    scenario_name,
    num_envs: int = 32,
    device: str = "cpu",
    continuous_actions: bool = True,
    wrapper: Optional[
        Wrapper
    ] = None,  # One of: None, vmas.Wrapper.RLLIB, and vmas.Wrapper.GYM
    max_steps: Optional[int] = None,
    seed: Optional[int] = None,
    **kwargs,
):
    # load scenario from script
    if not scenario_name.endswith(".py"):
        scenario_name += ".py"
    scenario = scenarios.load(scenario_name).Scenario()
    env = Environment(
        scenario,
        num_envs=num_envs,
        device=device,
        continuous_actions=continuous_actions,
        max_steps=max_steps,
        seed=seed,
        **kwargs,
    )

    return wrapper.get_env(env) if wrapper is not None else env



#  Copyright (c) 2022.
#  ProrokLab (https://www.proroklab.org/)
#  All rights reserved.

import torch

from vmas import render_interactively
from vmas.simulator.core import Agent, World, Sphere, Landmark, Box, Line
from vmas.simulator.joints import Joint
from vmas.simulator.scenario import BaseScenario
from vmas.simulator.utils import Color


class Scenario(BaseScenario):
    def make_world(self, batch_dim: int, device: torch.device, **kwargs):
        self.n_agents = kwargs.get("n_agents", 5)
        self.with_joints = kwargs.get("joints", True)

        self.agent_dist = 0.1
        self.agent_radius = 0.04

        # Make world
        world = World(
            batch_dim, device, dt=0.1, drag=0.25, substeps=5, collision_force=500
        )
        # Add agents
        for i in range(self.n_agents):
            agent = Agent(
                name=f"agent {i}",
                shape=Sphere(radius=self.agent_radius),
                u_multiplier=0.7,
                rotatable=True,
            )
            world.add_agent(agent)
        if self.with_joints:
            # Add joints
            for i in range(self.n_agents - 1):
                joint = Joint(
                    world.agents[i],
                    world.agents[i + 1],
                    anchor_a=(1, 0),
                    anchor_b=(-1, 0),
                    dist=self.agent_dist,
                    rotate_a=True,
                    rotate_b=True,
                    collidable=True,
                    width=0,
                    mass=1,
                )
                world.add_joint(joint)
            landmark = Landmark(
                name="joined landmark",
                collide=True,
                movable=True,
                rotatable=True,
                shape=Box(length=self.agent_radius * 2, width=0.3),
                color=Color.GREEN,
            )
            world.add_landmark(landmark)
            joint = Joint(
                world.agents[-1],
                landmark,
                anchor_a=(1, 0),
                anchor_b=(-1, 0),
                dist=self.agent_dist,
                rotate_a=True,
                rotate_b=True,
                collidable=True,
                width=0,
                mass=1,
            )
            world.add_joint(joint)

        # Add landmarks
        for i in range(5):
            landmark = Landmark(
                name=f"landmark {i}",
                collide=True,
                movable=True,
                rotatable=True,
                shape=Box(length=0.3, width=0.1),
                color=Color.RED,
                # collision_filter=lambda e: False
                # if isinstance(e.shape, Box) and e.name != "joined landmark"
                # else True,
            )
            world.add_landmark(landmark)
        floor = Landmark(
            name="floor",
            collide=True,
            movable=False,
            shape=Line(length=2),
            color=Color.BLACK,
        )
        world.add_landmark(floor)

        return world

    def reset_world_at(self, env_index: int = None):
        for i, agent in enumerate(
            self.world.agents + [self.world.landmarks[self.n_agents - 1]]
        ):
            agent.set_pos(
                torch.tensor(
                    [-0.2 + (self.agent_dist + 2 * self.agent_radius) * i, 1.0],
                    dtype=torch.float32,
                    device=self.world.device,
                ),
                batch_index=env_index,
            )
        for i, landmark in enumerate(
            self.world.landmarks[(self.n_agents + 1) if self.with_joints else 0 : -1]
        ):
            landmark.set_pos(
                torch.tensor(
                    [0.2 if i % 2 else -0.2, 0.6 - 0.3 * i],
                    dtype=torch.float32,
                    device=self.world.device,
                ),
                batch_index=env_index,
            )
            landmark.set_rot(
                torch.tensor(
                    [torch.pi / 4 if i % 2 else -torch.pi / 4],
                    dtype=torch.float32,
                    device=self.world.device,
                ),
                batch_index=env_index,
            )
        floor = self.world.landmarks[-1]
        floor.set_pos(
            torch.tensor(
                [0, -1],
                dtype=torch.float32,
                device=self.world.device,
            ),
            batch_index=env_index,
        )

    def reward(self, agent: Agent):
        dist2 = torch.linalg.vector_norm(
            agent.state.pos - self.world.landmarks[-1].state.pos, dim=1
        )
        return -dist2

    def observation(self, agent: Agent):
        # get positions of all entities in this agent's reference frame
        return torch.cat(
            [agent.state.pos, agent.state.vel]
            + [
                landmark.state.pos - agent.state.pos
                for landmark in self.world.landmarks
            ],
            dim=-1,
        )


if __name__ == "__main__":
    render_interactively(
        __file__,
        control_two_agents=True,
        n_agents=5,
        joints=True,
    )



#  Copyright (c) 2022-2023.
#  ProrokLab (https://www.proroklab.org/)
#  All rights reserved.
"""
Use this script to interactively play with scenarios

You can change agent by pressing TAB
You can reset the environment by pressing R
You can move agents with the arrow keys
If you have more than 1 agent, you can control another one with W,A,S,D
and switch the agent with these controls using LSHIFT
"""
from operator import add

import numpy as np

from vmas.make_env import make_env
from vmas.simulator.environment import Wrapper
from vmas.simulator.environment.gym import GymWrapper
from vmas.simulator.utils import save_video

N_TEXT_LINES_INTERACTIVE = 6


class InteractiveEnv:
    """
    Use this script to interactively play with scenarios

    You can change agent by pressing TAB
    You can reset the environment by pressing R
    You can move agents with the arrow keys
    If you have more than 1 agent, you can control another one with W,A,S,D
    and switch the agent with these controls using LSHIFT
    """

    def __init__(
        self,
        env: GymWrapper,
        control_two_agents: bool = False,
        display_info: bool = True,
        save_render: bool = False,
        render_name: str = "interactive",
    ):

        self.env = env
        self.control_two_agents = control_two_agents
        # hard-coded keyboard events
        self.current_agent_index = 0
        self.current_agent_index2 = 1
        self.n_agents = self.env.unwrapped().n_agents
        self.continuous = self.env.unwrapped().continuous_actions
        self.reset = False
        self.keys = np.array([0.0, 0.0, 0.0, 0.0])  # up, down, left, right
        self.keys2 = np.array([0.0, 0.0, 0.0, 0.0])  # up, down, left, right
        self.u = 0 if not self.continuous else (0.0, 0.0)
        self.u2 = 0 if not self.continuous else (0.0, 0.0)
        self.frame_list = []
        self.display_info = display_info
        self.save_render = save_render
        self.render_name = render_name

        if self.control_two_agents:
            assert (
                self.n_agents >= 2
            ), "Control_two_agents is true but not enough agents in scenario"

        self.env.render()
        self._init_text()
        self.env.unwrapped().viewer.window.on_key_press = self._key_press
        self.env.unwrapped().viewer.window.on_key_release = self._key_release

        self._cycle()

    def _increment_selected_agent_index(self, index: int):
        index += 1
        if index == self.n_agents:
            index = 0
        return index

    def _cycle(self):
        total_rew = [0] * self.n_agents
        while True:
            if self.reset:
                if self.save_render:
                    save_video(
                        self.render_name,
                        self.frame_list,
                        fps=1 / self.env.env.world.dt,
                    )
                self.env.reset()
                self.reset = False
                total_rew = [0] * self.n_agents

            if self.continuous:
                action_list = [(0.0, 0.0)] * self.n_agents
            else:
                action_list = [0] * self.n_agents
            action_list[self.current_agent_index] = self.u
            if self.n_agents > 1 and self.control_two_agents:
                action_list[self.current_agent_index2] = self.u2
            obs, rew, done, info = self.env.step(action_list)

            if self.display_info:
                obs[self.current_agent_index] = np.around(
                    obs[self.current_agent_index].cpu().tolist(), decimals=2
                )
                len_obs = len(obs[self.current_agent_index])
                message = f"\t\t{obs[self.current_agent_index][len_obs//2:]}"
                self._write_values(self.text_idx, message)
                message = f"Obs: {obs[self.current_agent_index][:len_obs//2]}"
                self._write_values(self.text_idx + 1, message)

                message = f"Rew: {round(rew[self.current_agent_index],3)}"
                self._write_values(self.text_idx + 2, message)

                total_rew = list(map(add, total_rew, rew))
                message = f"Total rew: {round(total_rew[self.current_agent_index], 3)}"
                self._write_values(self.text_idx + 3, message)

                message = f"Done: {done}"
                self._write_values(self.text_idx + 4, message)

                message = f"Selected: {self.env.unwrapped().agents[self.current_agent_index].name}"
                self._write_values(self.text_idx + 5, message)

            self.frame_list.append(
                self.env.render(
                    mode="rgb_array",
                    visualize_when_rgb=True,
                )
            )

            if done:
                self.reset = True

    def _init_text(self):
        from vmas.simulator import rendering

        try:
            self.text_idx = len(self.env.unwrapped().viewer.text_lines)
        except AttributeError:
            self.text_idx = 0

        for i in range(N_TEXT_LINES_INTERACTIVE):
            text_line = rendering.TextLine(
                self.env.unwrapped().viewer.window, self.text_idx + i
            )
            self.env.unwrapped().viewer.text_lines.append(text_line)

    def _write_values(self, index: int, message: str, font_size: int = 15):
        self.env.unwrapped().viewer.text_lines[index].set_text(
            message, font_size=font_size
        )

    # keyboard event callbacks
    def _key_press(self, k, mod):
        from pyglet.window import key

        agent_range = self.env.unwrapped().agents[self.current_agent_index].u_range

        u = self.u
        u2 = self.u2
        if k == key.LEFT:
            self.keys[0] = agent_range
            u = 1
        elif k == key.RIGHT:
            self.keys[1] = agent_range
            u = 2
        elif k == key.DOWN:
            self.keys[2] = agent_range
            u = 3
        elif k == key.UP:
            self.keys[3] = agent_range
            u = 4
        elif k == key.TAB:
            self.current_agent_index = self._increment_selected_agent_index(
                self.current_agent_index
            )
            if self.control_two_agents:
                while self.current_agent_index == self.current_agent_index2:
                    self.current_agent_index = self._increment_selected_agent_index(
                        self.current_agent_index
                    )

        if self.control_two_agents:
            agent2_range = (
                self.env.unwrapped().agents[self.current_agent_index2].u_range
            )
            if k == key.A:
                self.keys2[0] = agent2_range
                u2 = 1
            elif k == key.D:
                self.keys2[1] = agent2_range
                u2 = 2
            elif k == key.S:
                self.keys2[2] = agent2_range
                u2 = 3
            elif k == key.W:
                self.keys2[3] = agent2_range
                u2 = 4
            elif k == key.LSHIFT:
                self.current_agent_index2 = self._increment_selected_agent_index(
                    self.current_agent_index2
                )
                while self.current_agent_index == self.current_agent_index2:
                    self.current_agent_index2 = self._increment_selected_agent_index(
                        self.current_agent_index2
                    )

        if k == key.R:
            self.reset = True

        if self.continuous:
            self.u = (self.keys[1] - self.keys[0], self.keys[3] - self.keys[2])
            self.u2 = (self.keys2[1] - self.keys2[0], self.keys2[3] - self.keys2[2])
        else:
            self.u = u
            self.u2 = u2

    def _key_release(self, k, mod):
        from pyglet.window import key

        if k == key.LEFT:
            self.keys[0] = 0
        elif k == key.RIGHT:
            self.keys[1] = 0
        elif k == key.DOWN:
            self.keys[2] = 0
        elif k == key.UP:
            self.keys[3] = 0

        elif k == key.A:
            self.keys2[0] = 0
        elif k == key.D:
            self.keys2[1] = 0
        elif k == key.S:
            self.keys2[2] = 0
        elif k == key.W:
            self.keys2[3] = 0

        elif k == key.R:
            self.reset = False

        if self.continuous:
            self.u = (self.keys[1] - self.keys[0], self.keys[3] - self.keys[2])
            self.u2 = (self.keys2[1] - self.keys2[0], self.keys2[3] - self.keys2[2])
        else:
            if np.sum(self.keys) == 1:
                self.u = np.argmax(self.keys) + 1
            else:
                self.u = 0
            if np.sum(self.keys2) == 1:
                self.u2 = np.argmax(self.keys2) + 1
            else:
                self.u2 = 0


def render_interactively(
    scenario_name: str,
    control_two_agents: bool = False,
    display_info: bool = True,
    save_render: bool = False,
    **kwargs,
):
    """
    Use this script to interactively play with scenarios

    You can change agent by pressing TAB
    You can reset the environment by pressing R
    You can move agents with the arrow keys
    If you have more than 1 agent, you can control another one with W,A,S,D
    and switch the agent with these controls using LSHIFT
    """
    if scenario_name.endswith(".py"):
        scenario_name = scenario_name[:-3]

    InteractiveEnv(
        make_env(
            scenario_name=scenario_name,
            num_envs=1,
            device="cpu",
            continuous_actions=True,
            wrapper=Wrapper.GYM,
            seed=0,
            # Environment specific variables
            **kwargs,
        ),
        control_two_agents=control_two_agents,
        display_info=display_info,
        save_render=save_render,
        render_name=scenario_name,
    )


if __name__ == "__main__":
    # Use this script to interactively play with scenarios
    #
    # You can change agent by pressing TAB
    # You can reset the environment by pressing R
    # You can move agents with the arrow keys
    # If you have more than 1 agent, you can control another one with W,A,S,D
    # and switch the agent with these controls using LSHIFT

    scenario_name = "waterfall"

    # Scenario specific variables

    render_interactively(
        scenario_name, control_two_agents=True, save_render=False, display_info=True
    )



#  Copyright (c) 2022.
#  ProrokLab (https://www.proroklab.org/)
#  All rights reserved.

import torch

from vmas.simulator.core import World, Agent, Landmark, Sphere
from vmas.simulator.scenario import BaseScenario


class Scenario(BaseScenario):
    def make_world(self, batch_dim: int, device: torch.device, **kwargs):

        world = World(batch_dim=batch_dim, device=device, dim_c=3)
        # set any world properties first
        num_agents = 2
        num_landmarks = 3

        # Add agents
        for i in range(num_agents):
            speaker = True if i == 0 else False
            agent = Agent(
                name=f"agent {i}",
                collide=False,
                movable=False if speaker else True,
                silent=False if speaker else True,
                shape=Sphere(radius=0.075),
            )
            world.add_agent(agent)
        # Add landmarks
        for i in range(num_landmarks):
            landmark = Landmark(
                name=f"landmark {i}", collide=False, shape=Sphere(radius=0.04)
            )
            world.add_landmark(landmark)

        return world

    def reset_world_at(self, env_index: int = None):
        if env_index is None:
            # assign goals to agents
            for agent in self.world.agents:
                agent.goal_a = None
                agent.goal_b = None
            # want listener to go to the goal landmark
            self.world.agents[0].goal_a = self.world.agents[1]
            self.world.agents[0].goal_b = self.world.landmarks[
                torch.randint(0, len(self.world.landmarks), (1,)).item()
            ]
            # random properties for agents
            for i, agent in enumerate(self.world.agents):
                agent.color = torch.tensor(
                    [0.25, 0.25, 0.25], device=self.world.device, dtype=torch.float32
                )
            # random properties for landmarks
            self.world.landmarks[0].color = torch.tensor(
                [0.65, 0.15, 0.15], device=self.world.device, dtype=torch.float32
            )
            self.world.landmarks[1].color = torch.tensor(
                [0.15, 0.65, 0.15], device=self.world.device, dtype=torch.float32
            )
            self.world.landmarks[2].color = torch.tensor(
                [0.15, 0.15, 0.65], device=self.world.device, dtype=torch.float32
            )
            # special colors for goals
            self.world.agents[0].goal_a.color = self.world.agents[
                0
            ].goal_b.color + torch.tensor(
                [0.45, 0.45, 0.45], device=self.world.device, dtype=torch.float32
            )

        # set random initial states
        for agent in self.world.agents:
            agent.set_pos(
                2
                * torch.rand(
                    self.world.dim_p, device=self.world.device, dtype=torch.float32
                )
                - 1,
                batch_index=env_index,
            )
        for landmark in self.world.landmarks:
            landmark.set_pos(
                2
                * torch.rand(
                    self.world.dim_p, device=self.world.device, dtype=torch.float32
                )
                - 1,
                batch_index=env_index,
            )

    def reward(self, agent: Agent):
        # squared distance from listener to landmark
        a = self.world.agents[0]
        dist2 = torch.sqrt(
            torch.sum(torch.square(a.goal_a.state.pos - a.goal_b.state.pos), dim=-1)
        )
        return -dist2

    def observation(self, agent):
        # goal color
        goal_color = torch.zeros(3, device=self.world.device, dtype=torch.float32)
        if agent.goal_b is not None:
            goal_color = agent.goal_b.color

        # get positions of all entities in this agent's reference frame
        entity_pos = []
        for entity in self.world.landmarks:
            entity_pos.append(entity.state.pos - agent.state.pos)

        # communication of all other agents
        comm = []
        for other in self.world.agents:
            if other is agent or (other.state.c is None):
                continue
            comm.append(other.state.c)

        # speaker
        if not agent.movable:
            return goal_color.repeat(self.world.batch_dim, 1)
        # listener
        if agent.silent:
            return torch.cat([agent.state.vel, *entity_pos, *comm], dim=-1)



#  Copyright (c) 2022-2023.
#  ProrokLab (https://www.proroklab.org/)
#  All rights reserved.
import random
from ctypes import byref
from typing import List, Tuple, Callable, Optional

import numpy as np
import torch
from gym import spaces
from torch import Tensor

import vmas.simulator.utils
from vmas.simulator.core import Agent, TorchVectorizedObject
from vmas.simulator.scenario import BaseScenario
from vmas.simulator.utils import VIEWER_MIN_ZOOM
from vmas.simulator.utils import X, Y, ALPHABET


# environment for all agents in the multiagent world
# currently code assumes that no agents will be created/destroyed at runtime!
class Environment(TorchVectorizedObject):
    metadata = {
        "render.modes": ["human", "rgb_array"],
        "runtime.vectorized": True,
    }

    def __init__(
        self,
        scenario: BaseScenario,
        num_envs: int = 32,
        device: str = "cpu",
        max_steps: Optional[int] = None,
        continuous_actions: bool = True,
        seed: Optional[int] = None,
        **kwargs,
    ):

        self.scenario = scenario
        self.num_envs = num_envs
        TorchVectorizedObject.__init__(self, num_envs, torch.device(device))
        self.world = self.scenario.env_make_world(self.num_envs, self.device, **kwargs)

        self.agents = self.world.policy_agents
        self.n_agents = len(self.agents)
        self.max_steps = max_steps
        self.continuous_actions = continuous_actions

        self.reset(seed=seed)

        # configure spaces
        self.action_space = spaces.Tuple(
            [self.get_action_space(agent) for agent in self.agents]
        )
        self.observation_space = spaces.Tuple(
            [
                spaces.Box(
                    low=-np.float32("inf"),
                    high=np.float32("inf"),
                    shape=(len(self.scenario.observation(agent)[0]),),
                    dtype=np.float32,
                )
                for agent in self.agents
            ]
        )

        # rendering
        self.render_geoms_xform = None
        self.render_geoms = None
        self.viewer = None
        self.headless = None
        self.visible_display = None

    def reset(self, seed: Optional[int] = None, return_info: bool = False):
        """
        Resets the environment in a vectorized way
        Returns observations for all envs and agents
        """
        if seed is not None:
            self.seed(seed)
        # reset world
        self.scenario.env_reset_world_at(env_index=None)
        self.steps = torch.zeros(self.num_envs, device=self.device)
        # record observations for each agent
        obs = []
        info = []
        for agent in self.agents:
            obs.append(self.scenario.observation(agent))
            if return_info:
                info.append(self.scenario.info(agent))
        return (obs, info) if return_info else obs

    def reset_at(self, index: int, return_info: bool = False):
        """
        Resets the environment at index
        Returns observations for all agents in that environment
        """
        self._check_batch_index(index)
        self.scenario.env_reset_world_at(index)
        self.steps[index] = 0
        obs = []
        info = []
        for agent in self.agents:
            obs.append(self.scenario.observation(agent)[index].unsqueeze(0))
            if return_info:
                info.append(
                    {
                        key: val[index].unsqueeze(0)
                        for key, val in self.scenario.info(agent).items()
                    }
                )
        return (obs, info) if return_info else obs

    def seed(self, seed=None):
        if seed is None:
            seed = 0
        torch.manual_seed(seed)
        np.random.seed(seed)
        random.seed(seed)
        return [seed]

    def step(self, actions: List):
        """Performs a vectorized step on all sub environments using `actions`.
        Args:
            actions: Is a list on len 'self.n_agents' of which each element is a torch.Tensor of shape
            '(self.num_envs, action_size_of_agent)'.
        Returns:
            obs: List on len 'self.n_agents' of which each element is a torch.Tensor
                 of shape '(self.num_envs, obs_size_of_agent)'
            rewards: List on len 'self.n_agents' of which each element is a torch.Tensor of shape '(self.num_envs)'
            dones: Tensor of len 'self.num_envs' of which each element is a bool
            infos : List on len 'self.n_agents' of which each element is a dictionary for which each key is a metric
                    and the value is a tensor of shape '(self.num_envs, metric_size_per_agent)'
        """
        assert (
            len(actions) == self.n_agents
        ), f"Expecting actions for {self.n_agents}, got {len(actions)} actions"
        for i in range(len(actions)):
            if not isinstance(actions[i], Tensor):
                actions[i] = torch.tensor(
                    actions[i], dtype=torch.float32, device=self.device
                )
            if len(actions[i].shape) == 1:
                actions[i].unsqueeze_(-1)
            assert (
                actions[i].shape[0] == self.num_envs
            ), f"Actions used in input of env must be of len {self.num_envs}, got {actions[i].shape[0]}"
            assert actions[i].shape[1] == self.get_agent_action_size(self.agents[i]), (
                f"Action for agent {self.agents[i].name} has shape {actions[i].shape[1]},"
                f" but should have shape {self.get_agent_action_size(self.agents[i])}"
            )

        # set action for each agent
        for i, agent in enumerate(self.agents):
            self._set_action(actions[i], agent)
        # Scenarios can define a custom action processor. This step takes care also of scripted agents automatically
        for agent in self.world.agents:
            self.scenario.env_process_action(agent)

        # advance world state
        self.world.step()

        obs = []
        rewards = []
        infos = []
        for agent in self.agents:
            rewards.append(self.scenario.reward(agent).clone())
            obs.append(self.scenario.observation(agent).clone())
            # A dictionary per agent
            infos.append(self.scenario.info(agent))

        dones = self.scenario.done().clone()

        self.steps += 1
        if self.max_steps is not None:
            dones += self.steps >= self.max_steps
        # print("\nStep results in unwrapped environment")
        # print(
        #     f"Actions len (n_agents): {len(actions)}, "
        #     f"actions[0] shape (num_envs, agent 0 action shape): {actions[0].shape}, "
        #     f"actions[0][0] (action agent 0 env 0): {actions[0][0]}"
        # )
        # print(
        #     f"Obs len (n_agents): {len(obs)}, "
        #     f"obs[0] shape (num_envs, agent 0 obs shape): {obs[0].shape}, obs[0][0] (obs agent 0 env 0): {obs[0][0]}"
        # )
        # print(
        #     f"Rewards len (n_agents): {len(rewards)}, rewards[0] shape (num_envs, 1): {rewards[0].shape}, "
        #     f"rewards[0][0] (agent 0 env 0): {rewards[0][0]}"
        # )
        # print(f"Dones len (n_envs): {len(dones)}, dones[0] (done env 0): {dones[0]}")
        # print(f"Info len (n_agents): {len(infos)}, info[0] (infos agent 0): {infos[0]}")
        return obs, rewards, dones, infos

    def get_agent_action_size(self, agent: Agent):
        return (
            self.world.dim_p
            + (1 if agent.action.u_rot_range != 0 else 0)
            + (self.world.dim_c if not agent.silent else 0)
            if self.continuous_actions
            else 1
            + (1 if agent.action.u_rot_range != 0 else 0)
            + (1 if not agent.silent else 0)
        )

    def get_action_space(self, agent: Agent):
        if self.continuous_actions:
            return spaces.Box(
                low=np.array(
                    [-np.float32(agent.u_range)] * self.world.dim_p
                    + [-np.float32(agent.u_rot_range)]
                    * (1 if agent.u_rot_range != 0 else 0)
                    + [0.0] * (self.world.dim_c if not agent.silent else 0)
                ),
                high=np.array(
                    [np.float32(agent.u_range)] * self.world.dim_p
                    + [np.float32(agent.u_rot_range)]
                    * (1 if agent.u_rot_range != 0 else 0)
                    + [1.0] * (self.world.dim_c if not agent.silent else 0)
                ),
                shape=(self.get_agent_action_size(agent),),
                dtype=np.float32,
            )
        else:
            if (self.world.dim_c == 0 or agent.silent) and agent.u_rot_range == 0.0:
                return spaces.Discrete(self.world.dim_p * 2 + 1)
            else:
                actions = (
                    [self.world.dim_p * 2 + 1]
                    + ([3] if agent.u_rot_range != 0 else [])
                    + (
                        [self.world.dim_c]
                        if not agent.silent and self.world.dim_c != 0
                        else []
                    )
                )
                return spaces.MultiDiscrete(actions)

    def _check_discrete_action(self, action: Tensor, low: int, high: int, type: str):
        assert torch.all(
            torch.any(
                torch.arange(low, high, device=self.device).repeat(self.num_envs)
                == action,
                dim=-1,
            )
        ), f"Discrete {type} actions are out of bounds, allowed int range [{low},{high})"

    # set env action for a particular agent
    def _set_action(self, action, agent):
        action = action.clone().to(self.device)
        agent.action.u = torch.zeros(
            self.batch_dim, self.world.dim_p, device=self.device, dtype=torch.float32
        )

        assert action.shape[1] == self.get_agent_action_size(agent), (
            f"Agent {agent.name} has wrong action size, got {action.shape[1]}, "
            f"expected {self.get_agent_action_size(agent)}"
        )
        action_index = 0

        if self.continuous_actions:
            physical_action = action[:, action_index : action_index + self.world.dim_p]
            action_index += self.world.dim_p
            assert not torch.any(
                torch.abs(physical_action) > agent.u_range
            ), f"Physical actions of agent {agent.name} are out of its range {agent.u_range}"

            agent.action.u = physical_action.to(torch.float32)
        else:
            physical_action = action[:, action_index].unsqueeze(-1)
            action_index += 1
            self._check_discrete_action(
                physical_action,
                low=0,
                high=self.world.dim_p * 2 + 1,
                type="physical",
            )

            arr1 = physical_action == 1
            arr2 = physical_action == 2
            arr3 = physical_action == 3
            arr4 = physical_action == 4

            disc_action_value = agent.u_range

            agent.action.u[:, X] -= disc_action_value * arr1.squeeze(-1)
            agent.action.u[:, X] += disc_action_value * arr2.squeeze(-1)
            agent.action.u[:, Y] -= disc_action_value * arr3.squeeze(-1)
            agent.action.u[:, Y] += disc_action_value * arr4.squeeze(-1)

        agent.action.u *= agent.u_multiplier
        if agent.u_rot_range != 0:
            agent.action.u_rot = torch.zeros(
                self.batch_dim, 1, device=self.device, dtype=torch.float32
            )
            if self.continuous_actions:
                physical_action = action[:, action_index]
                action_index += 1
                assert not torch.any(
                    torch.abs(physical_action) > agent.u_rot_range
                ), f"Physical rotation actions of agent {agent.name} are out of its range {agent.u_rot_range}"

                agent.action.u_rot = physical_action.to(torch.float32)
            else:
                physical_action = action[:, action_index].unsqueeze(-1)
                action_index += 1
                self._check_discrete_action(
                    physical_action,
                    low=0,
                    high=3,
                    type="rotation",
                )

                arr1 = physical_action == 1
                arr2 = physical_action == 2

                disc_action_value = agent.u_rot_range

                agent.action.u_rot[:] -= disc_action_value * arr1.squeeze(-1)
                agent.action.u_rot[:] += disc_action_value * arr2.squeeze(-1)

            agent.action.u_rot *= agent.u_rot_multiplier
        if self.world.dim_c > 0 and not agent.silent:
            if not self.continuous_actions:
                comm_action = action[:, action_index:]
                self._check_discrete_action(
                    comm_action, 0, self.world.dim_c, "communication"
                )
                comm_action = comm_action.long()
                agent.action.c = torch.zeros(
                    self.num_envs,
                    self.world.dim_c,
                    device=self.device,
                    dtype=torch.float32,
                )
                # Discrete to one-hot
                agent.action.c.scatter_(1, comm_action, 1)
            else:
                comm_action = action[:, action_index:]
                assert not torch.any(comm_action > 1) and not torch.any(
                    comm_action < 0
                ), "Comm actions are out of range [0,1]"
                agent.action.c = comm_action

    def render(
        self,
        mode="human",
        env_index=0,
        agent_index_focus: int = None,
        visualize_when_rgb: bool = False,
        plot_position_function: Callable[[Tuple[float, float]], float] = None,
        plot_position_function_precision: float = 0.05,
        plot_position_function_range: float = 1,
    ):
        """
        Render function for environment using pyglet

        On servers use mode="rgb_array" and set
        ```
        export DISPLAY=':99.0'
        Xvfb :99 -screen 0 1400x900x24 > /dev/null 2>&1 &
        ```

        :param mode: One of human or rgb_array
        :param env_index: Index of the environment to render
        :param agent_index_focus: If specified the camera will stay on the agent with this index.
                                  If None, the camera will stay in the center and zoom out to contain all agents
        :param visualize_when_rgb: Also run human visualization when mode=="rgb_array"
        :param plot_position_function: A function to plot under the rendering. This function takes
         the position (x,y) as input and outputs a transparency alpha value
        :param plot_position_function_precision: The precision to use for plotting the function
        :param plot_position_function_range: The position range to plot the function in
        :return: Rgb array or None, depending on the mode
        """
        self._check_batch_index(env_index)
        assert (
            mode in self.metadata["render.modes"]
        ), f"Invalid mode {mode} received, allowed modes: {self.metadata['render.modes']}"
        if agent_index_focus is not None:
            assert 0 <= agent_index_focus < self.n_agents, (
                f"Agent focus in rendering should be a valid agent index"
                f" between 0 and {self.n_agents}, got {agent_index_focus}"
            )
        shared_viewer = agent_index_focus is None
        aspect_ratio = self.scenario.viewer_size[X] / self.scenario.viewer_size[Y]

        headless = mode == "rgb_array" and not visualize_when_rgb
        # First time rendering
        if self.visible_display is None:
            self.visible_display = not headless
            self.headless = headless
        # All other times headless should be the same
        else:
            assert self.visible_display is not headless

        # First time rendering
        if self.viewer is None:
            try:
                import pyglet
            except ImportError:
                raise ImportError(
                    "Cannot import pyg;et: you can install pyglet directly via 'pip install pyglet'."
                )

            try:
                # Try to use EGL
                pyglet.lib.load_library("EGL")

                # Only if we have GPUs
                from pyglet.libs.egl import egl
                from pyglet.libs.egl import eglext

                num_devices = egl.EGLint()
                eglext.eglQueryDevicesEXT(0, None, byref(num_devices))
                assert num_devices.value > 0

            except (ImportError, AssertionError):
                self.headless = False
            pyglet.options["headless"] = self.headless

            self._init_rendering()

        # Render comm messages
        if self.world.dim_c > 0:
            idx = 0
            for agent in self.world.agents:
                if agent.silent:
                    continue
                assert (
                    agent.state.c is not None
                ), "Agent has no comm state but it should"
                if self.continuous_actions:
                    word = (
                        "["
                        + ",".join([f"{comm:.2f}" for comm in agent.state.c[env_index]])
                        + "]"
                    )
                else:
                    word = ALPHABET[torch.argmax(agent.state.c[env_index]).item()]

                message = agent.name + " sends " + word + "   "
                self.viewer.text_lines[idx].set_text(message)
                idx += 1

        zoom = max(VIEWER_MIN_ZOOM, self.scenario.viewer_zoom)

        if aspect_ratio < 1:
            cam_range = torch.tensor([zoom, zoom / aspect_ratio], device=self.device)
        else:
            cam_range = torch.tensor([zoom * aspect_ratio, zoom], device=self.device)

        if shared_viewer:
            # zoom out to fit everyone
            all_poses = torch.stack(
                [agent.state.pos[env_index] for agent in self.world.agents], dim=0
            )
            max_agent_radius = np.max(
                [agent.shape.circumscribed_radius() for agent in self.world.agents]
            )
            viewer_size_fit = (
                torch.stack(
                    [
                        torch.max(torch.abs(all_poses[:, X])),
                        torch.max(torch.abs(all_poses[:, Y])),
                    ]
                )
                + 2 * max_agent_radius
            )

            viewer_size = torch.maximum(
                viewer_size_fit / cam_range, torch.tensor(zoom, device=self.device)
            )
            cam_range *= torch.max(viewer_size)
            self.viewer.set_bounds(
                -cam_range[X],
                cam_range[X],
                -cam_range[Y],
                cam_range[Y],
            )
        else:
            # update bounds to center around agent
            pos = self.agents[agent_index_focus].state.pos[env_index]
            self.viewer.set_bounds(
                pos[X] - cam_range[X],
                pos[X] + cam_range[X],
                pos[Y] - cam_range[Y],
                pos[Y] + cam_range[Y],
            )
        from vmas.simulator.rendering import Grid

        if plot_position_function is not None:
            self.plot_function(
                plot_position_function,
                precision=plot_position_function_precision,
                plot_range=plot_position_function_range,
            )

        if self.scenario.plot_grid:
            grid = Grid(spacing=self.scenario.grid_spacing)
            grid.set_color(*vmas.simulator.utils.Color.BLACK.value, alpha=0.3)
            self.viewer.add_onetime(grid)

        for geom in self.scenario.extra_render(env_index):
            self.viewer.add_onetime(geom)

        for entity in self.world.entities:
            [
                self.viewer.add_onetime(geom)
                for geom in entity.render(env_index=env_index)
            ]

        # render to display or array
        return self.viewer.render(return_rgb_array=mode == "rgb_array")

    def plot_function(
        self,
        f: Callable[[Tuple[float, float]], float],
        precision: float,
        plot_range: float,
    ):
        from vmas.simulator.rendering import Transform, make_polygon

        l, r, t, b = (
            0,
            precision,
            precision,
            0,
        )
        poly_points = [(l, b), (l, t), (r, t), (r, b)]

        points = np.arange(-plot_range, plot_range, precision)
        for x in points:
            for y in points:
                alpha = f(x, y)
                box = make_polygon(poly_points, draw_border=False)
                transform = Transform()
                transform.set_translation(x, y)
                box.add_attr(transform)
                box.set_color(*vmas.simulator.utils.Color.BLACK.value, alpha=alpha)
                self.viewer.add_onetime(box)

    def _init_rendering(self):
        from vmas.simulator import rendering

        self.viewer = rendering.Viewer(
            *self.scenario.viewer_size, visible=self.visible_display
        )

        idx = 0
        if self.world.dim_c > 0:
            self.viewer.text_lines = []
            for agent in self.world.agents:
                if not agent.silent:
                    text_line = rendering.TextLine(self.viewer.window, idx)
                    self.viewer.text_lines.append(text_line)
                    idx += 1



#  Copyright (c) 2022.
#  ProrokLab (https://www.proroklab.org/)
#  All rights reserved.

import torch

from vmas import render_interactively
from vmas.simulator.core import Agent, Landmark, World, Line, Sphere
from vmas.simulator.heuristic_policy import BaseHeuristicPolicy
from vmas.simulator.scenario import BaseScenario
from vmas.simulator.utils import Color, TorchUtils


class Scenario(BaseScenario):
    def make_world(self, batch_dim: int, device: torch.device, **kwargs):
        n_agents = kwargs.get("n_agents", 4)
        self.line_length = kwargs.get("line_length", 2)
        line_mass = kwargs.get("line_mass", 30)
        self.desired_velocity = kwargs.get("desired_velocity", 0.05)

        # Make world
        world = World(batch_dim, device)
        # Add agents
        for i in range(n_agents):
            # Constraint: all agents have same action range and multiplier
            agent = Agent(name=f"agent {i}", u_multiplier=0.6, shape=Sphere(0.03))
            world.add_agent(agent)
        # Add landmarks
        self.line = Landmark(
            name="line",
            collide=True,
            rotatable=True,
            shape=Line(length=self.line_length),
            mass=line_mass,
            color=Color.BLACK,
        )
        world.add_landmark(self.line)
        center = Landmark(
            name="center",
            shape=Sphere(radius=0.02),
            collide=False,
            color=Color.BLACK,
        )
        world.add_landmark(center)

        return world

    def reset_world_at(self, env_index: int = None):
        for agent in self.world.agents:
            # Random pos between -1 and 1
            agent.set_pos(
                torch.zeros(
                    (1, self.world.dim_p)
                    if env_index is not None
                    else (self.world.batch_dim, self.world.dim_p),
                    device=self.world.device,
                    dtype=torch.float32,
                ).uniform_(
                    -1.0,
                    1.0,
                ),
                batch_index=env_index,
            )

        self.line.set_rot(
            torch.zeros(
                (1, 1) if env_index is not None else (self.world.batch_dim, 1),
                device=self.world.device,
                dtype=torch.float32,
            ).uniform_(
                -torch.pi / 2,
                torch.pi / 2,
            ),
            batch_index=env_index,
        )

    def reward(self, agent: Agent):
        is_first = agent == self.world.agents[0]

        if is_first:
            self.rew = (self.line.state.ang_vel.abs() - self.desired_velocity).abs()

        return -self.rew

    def observation(self, agent: Agent):

        line_end_1 = torch.cat(
            [
                (self.line_length / 2) * torch.cos(self.line.state.rot),
                (self.line_length / 2) * torch.sin(self.line.state.rot),
            ],
            dim=1,
        )
        line_end_2 = -line_end_1

        return torch.cat(
            [
                agent.state.pos,
                agent.state.vel,
                self.line.state.pos - agent.state.pos,
                line_end_1 - agent.state.pos,
                line_end_2 - agent.state.pos,
                self.line.state.rot % torch.pi,
                self.line.state.ang_vel.abs(),
                (self.line.state.ang_vel.abs() - self.desired_velocity).abs(),
            ],
            dim=-1,
        )


class HeuristicPolicy(BaseHeuristicPolicy):
    def compute_action(self, observation: torch.Tensor, u_range: float) -> torch.Tensor:
        assert self.continuous_actions is True, "Heuristic for continuous actions only"

        index_line_extrema = 6

        pos_agent = observation[:, :2]
        pos_end2_agent = observation[:, index_line_extrema + 2 : index_line_extrema + 4]

        pos_end2 = pos_end2_agent + pos_agent

        pos_end2_shifted = TorchUtils.rotate_vector(
            pos_end2, torch.tensor(torch.pi / 4, device=observation.device)
        )

        pos_end2_shifted_agent = pos_end2_shifted - pos_agent

        action_agent = torch.clamp(
            pos_end2_shifted_agent,
            min=-u_range,
            max=u_range,
        )

        return action_agent


if __name__ == "__main__":
    render_interactively(
        __file__,
        control_two_agents=True,
        desired_velocity=0.05,
        n_agents=4,
        line_length=2,
        line_mass=30,
    )
